"""
test_gaussian_gjf_input.py, Geoffrey Weal, 23/5/24

This will test the read input for gjf files

Notes
-----
* 23/5/24 -> Wrote test for testing the charge and mutliplicity inputs.
"""
import warnings
from io import StringIO

import pytest

from ase.calculators.calculator import compare_atoms
from ase.io import ParseError, read

# -----------------------------------------------------------
# The following strings give the components of an example gif file
# for two methane molecules.


# The input line of the gjf file.
ORIGINAL_GJF_FILE_PREFIX = """%chk=Preview_s6ozz.chk
# opt freq wb97xd/cc-pvqz geom=connectivity

Title Card Required

"""


# Provide an example of two methane molecules.
ORIGINAL_GJF_FILE_SUFFIX = """
 C                  0.96604221    0.36299765    0.00000000
 H                  1.32269663   -0.64581235    0.00000000
 H                  1.32271505    0.86739584    0.87365150
 H                  1.32271505    0.86739584   -0.87365150
 H                 -0.10395779    0.36301084    0.00000000
 C                 -4.67798621   -1.41686181    0.00000000
 H                 -4.32133178   -2.42567181    0.00000000
 H                 -4.32131337   -0.91246362    0.87365150
 H                 -4.32131337   -0.91246362   -0.87365150
 H                 -5.74798621   -1.41684862    0.00000000"""


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


# Provide an example of two methane molecules, which have been identified
# as different fragment.
FRAGMENT_GJF_FILE_SUFFIX = """
 C(Fragment=1)                  0.96604221    0.36299765    0.00000000
 H(Fragment=1)                  1.32269663   -0.64581235    0.00000000
 H(Fragment=1)                  1.32271505    0.86739584    0.87365150
 H(Fragment=1)                  1.32271505    0.86739584   -0.87365150
 H(Fragment=1)                 -0.10395779    0.36301084    0.00000000
 C(Fragment=2)                 -4.67798621   -1.41686181    0.00000000
 H(Fragment=2)                 -4.32133178   -2.42567181    0.00000000
 H(Fragment=2)                 -4.32131337   -0.91246362    0.87365150
 H(Fragment=2)                 -4.32131337   -0.91246362   -0.87365150
 H(Fragment=2)                 -5.74798621   -1.41684862    0.00000000"""


# -----------------------------------------------------------


# This is the original charge and multiplicity format.
ORIGINAL_CHARGE_MULTIPLICITY = "0 1"


# This is the fragment charge and multiplicity format that should work.
FRAGMENT_CHARGE_MULTIPLICITY_SUCCESS_1 = '0 1'
FRAGMENT_CHARGE_MULTIPLICITY_SUCCESS_2 = '0 1 0 1 0 1'


# This is the fragment charge and multiplicity format that should work, even
# if it wont make sense to Gaussian. But for ASE, this is fine, as ASE
# currently only reads in the total charge and the total multiplicity for
# the total system, which is given by the first two digits in this line.
FRAGMENT_CHARGE_MULTIPLICITY_SUCCESS_3 = '0 1 0 1'
FRAGMENT_CHARGE_MULTIPLICITY_SUCCESS_4 = '0 1 0 1 0 1 0 1'


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


# This is the original Gaussian file that should fail.
ORIGINAL_CHARGE_MULTIPLICITY_FAIL_1 = ' '
ORIGINAL_CHARGE_MULTIPLICITY_FAIL_2 = '0 '


# These are the fragment charge and multiplcities that should fail.
FRAGMENT_CHARGE_MULTIPLICITY_FAIL_1 = '0 '
FRAGMENT_CHARGE_MULTIPLICITY_FAIL_2 = '0 1 0'
FRAGMENT_CHARGE_MULTIPLICITY_FAIL_3 = '0 1 0 1 0'
FRAGMENT_CHARGE_MULTIPLICITY_FAIL_4 = '0 1 0 1 0 1 0'


# -----------------------------------------------------------


def _quiet_parse(input_string):
    """Contains the code for reading in the Gaussian input file into ASE."""
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        return read(StringIO(input_string), format="gaussian-in")


@pytest.fixture()
def reference_system():
    """This is the reference system to compare tests to."""
    return _quiet_parse(ORIGINAL_GJF_FILE_PREFIX +
                        ORIGINAL_CHARGE_MULTIPLICITY +
                        ORIGINAL_GJF_FILE_SUFFIX)


# -----------------------------------------------------------


def test_original_gjf_success(reference_system):
    """Test to make sure an example gjf file that has been written
    correctly is being read in correctly."""
    atoms = _quiet_parse(ORIGINAL_GJF_FILE_PREFIX +
                         ORIGINAL_CHARGE_MULTIPLICITY +
                         ORIGINAL_GJF_FILE_SUFFIX)
    assert compare_atoms(atoms, reference_system) == []


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


def test_fragment_gjf_success_1(reference_system):
    """Test to make sure the first example gjf file that has been written
    correctly and contains fragments is being read in correctly."""
    atoms = _quiet_parse(ORIGINAL_GJF_FILE_PREFIX +
                         FRAGMENT_CHARGE_MULTIPLICITY_SUCCESS_1 +
                         FRAGMENT_GJF_FILE_SUFFIX)
    assert compare_atoms(atoms, reference_system) == []


def test_fragment_gjf_success_2(reference_system):
    """Test to make sure the second example gjf file that has been written
    correctly and contains fragments is being read in correctly."""
    atoms = _quiet_parse(ORIGINAL_GJF_FILE_PREFIX +
                         FRAGMENT_CHARGE_MULTIPLICITY_SUCCESS_2 +
                         FRAGMENT_GJF_FILE_SUFFIX)
    assert compare_atoms(atoms, reference_system) == []


def test_fragment_gjf_success_3(reference_system):
    """Test to make sure the third example gjf file that has been written
    correctly and contains fragments is being read in correctly."""
    atoms = _quiet_parse(ORIGINAL_GJF_FILE_PREFIX +
                         FRAGMENT_CHARGE_MULTIPLICITY_SUCCESS_3 +
                         FRAGMENT_GJF_FILE_SUFFIX)
    assert compare_atoms(atoms, reference_system) == []


def test_fragment_gjf_success_4(reference_system):
    """Test to make sure the fourth example gjf file that has been written
    correctly and contains fragments is being read in correctly."""
    atoms = _quiet_parse(ORIGINAL_GJF_FILE_PREFIX +
                         FRAGMENT_CHARGE_MULTIPLICITY_SUCCESS_4 +
                         FRAGMENT_GJF_FILE_SUFFIX)
    assert compare_atoms(atoms, reference_system) == []


# -----------------------------------------------------------


def test_original_gjf_fail_1():
    """Test to make sure the first example gjf file that is expected to fail
    does infact fail based on the ParseError."""
    with pytest.raises(ParseError):
        _quiet_parse(ORIGINAL_GJF_FILE_PREFIX +
                     ORIGINAL_CHARGE_MULTIPLICITY_FAIL_1 +
                     ORIGINAL_GJF_FILE_SUFFIX)


def test_original_gjf_fail_2():
    """Test to make sure the second example gjf file that is expected to fail
    does infact fail based on the ParseError."""
    with pytest.raises(ParseError):
        _quiet_parse(ORIGINAL_GJF_FILE_PREFIX +
                     ORIGINAL_CHARGE_MULTIPLICITY_FAIL_2 +
                     ORIGINAL_GJF_FILE_SUFFIX)


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


def test_fragment_gjf_fail_1():
    """Test to make sure the first example gjf file that is expected to fail
    and contains fragments does infact fail based on the ParseError."""
    with pytest.raises(ParseError):
        _quiet_parse(ORIGINAL_GJF_FILE_PREFIX +
                     FRAGMENT_CHARGE_MULTIPLICITY_FAIL_1 +
                     FRAGMENT_GJF_FILE_SUFFIX)


def test_fragment_gjf_fail_2():
    """Test to make sure the second example gjf file that is expected to fail
    and contains fragments does infact fail based on the ParseError."""
    with pytest.raises(ParseError):
        _quiet_parse(ORIGINAL_GJF_FILE_PREFIX +
                     FRAGMENT_CHARGE_MULTIPLICITY_FAIL_2 +
                     FRAGMENT_GJF_FILE_SUFFIX)


def test_fragment_gjf_fail_3():
    """Test to make sure the third example gjf file that is expected to fail
    and contains fragments does infact fail based on the ParseError."""
    with pytest.raises(ParseError):
        _quiet_parse(ORIGINAL_GJF_FILE_PREFIX +
                     FRAGMENT_CHARGE_MULTIPLICITY_FAIL_3 +
                     FRAGMENT_GJF_FILE_SUFFIX)


def test_fragment_gjf_fail_4():
    """Test to make sure the fourth example gjf file that is expected to fail
    and contains fragments does infact fail based on the ParseError."""
    with pytest.raises(ParseError):
        _quiet_parse(ORIGINAL_GJF_FILE_PREFIX +
                     FRAGMENT_CHARGE_MULTIPLICITY_FAIL_4 +
                     FRAGMENT_GJF_FILE_SUFFIX)


# -----------------------------------------------------------



import doctest
import importlib

import numpy as np
import pytest

module_names = """\
ase.atoms
ase.build.tools
ase.cell
ase.calculators.vasp.vasp
ase.collections.collection
ase.dft.kpoints
ase.eos
ase.formula
ase.geometry.cell
ase.geometry.geometry
ase.io.ulm
ase.lattice
ase.phasediagram
ase.spacegroup.spacegroup
ase.spacegroup.xtal
ase.symbols
""".split()


@pytest.mark.parametrize('modname', module_names)
def test_doctest(testdir, modname, recwarn):
    mod = importlib.import_module(modname)
    with np.printoptions(legacy='1.13'):
        doctest.testmod(mod, raise_on_error=True, verbose=True)
        nwarnings = len(recwarn.list)
        if modname == 'ase.phasediagram':
            assert nwarnings == 1
        else:
            assert nwarnings == 0



import numpy as np

import ase.units as units


def write_amber_coordinates(atoms, filename):
    from scipy.io import netcdf_file
    with netcdf_file(filename, 'w', mmap=False) as fout:
        _write_amber_coordinates(atoms, fout)


def read_amber_coordinates(filename):
    from scipy.io import netcdf_file
    with netcdf_file(filename, 'r', mmap=False) as fin:
        return _read_amber_coordinates(fin)


def _write_amber_coordinates(atoms, fout):
    fout.Conventions = 'AMBERRESTART'
    fout.ConventionVersion = "1.0"
    fout.title = 'Ase-generated-amber-restart-file'
    fout.application = "AMBER"
    fout.program = "ASE"
    fout.programVersion = "1.0"
    fout.createDimension('cell_spatial', 3)
    fout.createDimension('label', 5)
    fout.createDimension('cell_angular', 3)
    fout.createDimension('time', 1)
    time = fout.createVariable('time', 'd', ('time',))
    time.units = 'picosecond'
    time[0] = 0
    fout.createDimension('spatial', 3)
    spatial = fout.createVariable('spatial', 'c', ('spatial',))
    spatial[:] = np.asarray(list('xyz'))

    natom = len(atoms)
    fout.createDimension('atom', natom)
    coordinates = fout.createVariable('coordinates', 'd',
                                      ('atom', 'spatial'))
    coordinates.units = 'angstrom'
    coordinates[:] = atoms.get_positions()

    velocities = fout.createVariable('velocities', 'd',
                                     ('atom', 'spatial'))
    velocities.units = 'angstrom/picosecond'
    # Amber's units of time are 1/20.455 ps
    # Any other units are ignored in restart files, so these
    # are the only ones it is safe to print
    # See: http://ambermd.org/Questions/units.html
    # Apply conversion factor from ps:
    velocities.scale_factor = 20.455
    # get_velocities call returns velocities with units sqrt(eV/u)
    # so convert to Ang/ps
    factor = units.fs * 1000 / velocities.scale_factor
    velocities[:] = atoms.get_velocities() * factor

    fout.createVariable('cell_angular', 'c', ('cell_angular', 'label'))

    cell_spatial = fout.createVariable('cell_spatial', 'c', ('cell_spatial',))
    cell_spatial[:] = ['a', 'b', 'c']

    cell_lengths = fout.createVariable('cell_lengths', 'd', ('cell_spatial',))
    cell_lengths.units = 'angstrom'
    cell_lengths[:] = atoms.cell.lengths()

    if not atoms.cell.orthorhombic:
        raise ValueError('Non-orthorhombic cell not supported with amber')

    cell_angles = fout.createVariable('cell_angles', 'd', ('cell_angular',))
    cell_angles[:3] = 90.0
    cell_angles.units = 'degree'


def _read_amber_coordinates(fin):
    from ase import Atoms

    all_coordinates = fin.variables['coordinates'][:]
    get_last_frame = False
    if hasattr(all_coordinates, 'ndim'):
        if all_coordinates.ndim == 3:
            get_last_frame = True
    elif hasattr(all_coordinates, 'shape'):
        if len(all_coordinates.shape) == 3:
            get_last_frame = True
    if get_last_frame:
        all_coordinates = all_coordinates[-1]

    atoms = Atoms(positions=all_coordinates)

    if 'velocities' in fin.variables:
        all_velocities = fin.variables['velocities']
        if hasattr(all_velocities, 'units'):
            if all_velocities.units != b'angstrom/picosecond':
                raise Exception(
                    f'Unrecognised units {all_velocities.units}')
        if hasattr(all_velocities, 'scale_factor'):
            scale_factor = all_velocities.scale_factor
        else:
            scale_factor = 1.0
        all_velocities = all_velocities[:] * scale_factor
        all_velocities = all_velocities / (1000 * units.fs)
        if get_last_frame:
            all_velocities = all_velocities[-1]
        atoms.set_velocities(all_velocities)
    if 'cell_lengths' in fin.variables:
        all_abc = fin.variables['cell_lengths']
        if get_last_frame:
            all_abc = all_abc[-1]
        a, b, c = all_abc
        all_angles = fin.variables['cell_angles']
        if get_last_frame:
            all_angles = all_angles[-1]
        alpha, beta, gamma = all_angles

        if (all(angle > 89.99 for angle in [alpha, beta, gamma]) and
                all(angle < 90.01 for angle in [alpha, beta, gamma])):
            atoms.set_cell(
                np.array([[a, 0, 0],
                          [0, b, 0],
                          [0, 0, c]]))
            atoms.set_pbc(True)
        else:
            raise NotImplementedError('only rectangular cells are'
                                      ' implemented in ASE-AMBER')

    else:
        atoms.set_pbc(False)

    return atoms



import numpy as np
import pytest

from ase.build import bulk
from ase.calculators.eam import EAM
from ase.calculators.emt import EMT
from ase.calculators.lj import LennardJones
from ase.calculators.qmmm import ForceQMMM, RescaledCalculator
from ase.eos import EquationOfState
from ase.geometry import get_distances
from ase.neighborlist import neighbor_list
from ase.optimize import FIRE
from ase.units import GPa


@pytest.fixture()
def mm_calc():
    bulk_at = bulk("Cu", cubic=True)
    sigma = (bulk_at * 2).get_distance(0, 1) * (2. ** (-1. / 6))

    return LennardJones(sigma=sigma, epsilon=0.05)


@pytest.fixture()
def qm_calc():
    return EMT()


@pytest.fixture()
def bulk_at():
    bulk_at = bulk("Cu", cubic=True)

    return bulk_at


@pytest.mark.slow()
def test_qm_buffer_mask(qm_calc, mm_calc, bulk_at):
    """
    test number of atoms in qm_buffer_mask for
    spherical region in a fully periodic cell
    also tests that "region" array returns the same mapping
    """

    alat = bulk_at.cell[0, 0]
    N_cell_geom = 10
    at0 = bulk_at * N_cell_geom
    r = at0.get_distances(0, np.arange(len(at0)), mic=True)
    print("N_cell", N_cell_geom, 'N_MM', len(at0), "Size", N_cell_geom * alat)
    qm_rc = 5.37  # cutoff for EMC()

    for R_QM in [1.0e-3,  # one atom in the center
                 alat / np.sqrt(2.0) + 1.0e-3,  # should give 12 nearest
                                                # neighbours + central atom
                 alat + 1.0e-3]:  # should give 18 neighbours + central atom

        at = at0.copy()
        qm_mask = r < R_QM
        qm_buffer_mask_ref = r < 2 * qm_rc + R_QM
        # exclude atoms that are too far (in case of non spherical region)
        # this is the old way to do it
        _, r_qm_buffer = get_distances(at.positions[qm_buffer_mask_ref],
                                       at.positions[qm_mask], at.cell, at.pbc)
        updated_qm_buffer_mask = np.ones_like(at[qm_buffer_mask_ref])
        for i, r_qm in enumerate(r_qm_buffer):
            if r_qm.min() > 2 * qm_rc:
                updated_qm_buffer_mask[i] = False

        qm_buffer_mask_ref[qm_buffer_mask_ref] = updated_qm_buffer_mask
        '''
        print(f'R_QM             {R_QM}   N_QM        {qm_mask.sum()}')
        print(f'R_QM + buffer: {2 * qm_rc + R_QM:.2f}'
              f' N_QM_buffer {qm_buffer_mask_ref.sum()}')
        print(f'                     N_total:    {len(at)}')
        '''
        qmmm = ForceQMMM(at, qm_mask, qm_calc, mm_calc, buffer_width=2 * qm_rc)
        # build qm_buffer_mask and test it
        qmmm.initialize_qm_buffer_mask(at)
        # print(f'      Calculator N_QM_buffer:'
        #       f'    {qmmm.qm_buffer_mask.sum().sum()}')
        assert qmmm.qm_buffer_mask.sum() == qm_buffer_mask_ref.sum()
        # same test for qmmm.get_cluster()
        qm_cluster = qmmm.get_qm_cluster(at)
        assert len(qm_cluster) == qm_buffer_mask_ref.sum()
        # test region mappings
        region = qmmm.get_region_from_masks(at)
        qm_mask_region = region == "QM"
        assert qm_mask_region.sum() == qm_mask.sum()
        buffer_mask_region = region == "buffer"
        assert qm_mask_region.sum() + \
            buffer_mask_region.sum() == qm_buffer_mask_ref.sum()


def compare_qm_cell_and_pbc(qm_calc, mm_calc, bulk_at,
                            test_size=4,
                            expected_pbc=np.array([True, True, True]),
                            buffer_width=5 * 3.61):
    """
    test qm cell shape and choice of pbc:
    make a non-periodic pdc in a direction
    if qm_radius + buffer is larger than the original cell
    keep the periodic cell otherwise i. e. if cell[i, i] > qm_radius + buffer
    the scenario is controlled by the test_size used to create at0
    as well as buffer_width.
    If the size of the at0 is larger than the r_qm + buffer + vacuum
    the cell stays periodic and the size is the same is original
    otherwise cell is non-periodic and size is different.
    """

    alat = bulk_at.cell[0, 0]
    at0 = bulk_at * test_size
    r = at0.get_distances(0, np.arange(len(at0)), mic=True)
    # should give 12 nearest neighbours + atom in the center
    R_QM = alat / np.sqrt(2.0) + 1.0e-3
    qm_mask = r < R_QM

    qmmm = ForceQMMM(at0, qm_mask, qm_calc, mm_calc,
                     buffer_width=buffer_width)
    # equal to 1 alat
    # build qm_buffer_mask to build the cell
    qmmm.initialize_qm_buffer_mask(at0)
    qm_cluster = qmmm.get_qm_cluster(at0)

    # test if qm pbc match expected in qmmm.get_cluster()
    assert all(qm_cluster.pbc == expected_pbc)

    # test the cell size for qmmm.get_qm_cluster()
    if not all(expected_pbc):  # at least one F. avoid comparing empty arrays
        assert not all(qm_cluster.cell.lengths()[~expected_pbc] ==
                       at0.cell.lengths()[~expected_pbc])
    if any(expected_pbc):  # at least one T. avoid comparing empty arrays
        np.testing.assert_allclose(qm_cluster.cell.lengths()[expected_pbc],
                                   at0.cell.lengths()[expected_pbc])


@pytest.mark.parametrize("kwargs",
                         [  # test the case of a cluster in
                            # a fully periodic cell:
                            # fist qm_radius + buffer > cell,
                            # thus should give a cluster with pbc=[T, T, T]
                            # (qm cluster is the same as the original cell)'''
                            {"test_size": 4,
                             "expected_pbc": np.array([True, True, True]),
                             "buffer_width": 5 * 3.61},
                            # test the case of a spherical
                            # cluster in a fully periodic cell:
                            # fist qm_radius + buffer < cell,
                            # thus should give a cluster with pbc=[F, F, F]
                            # (qm cluster cell must be DIFFERENT
                            # form the original cell)
                            {"test_size": 4,
                             "expected_pbc": np.array([False, False, False]),
                             "buffer_width": 3.61},
                            # testing the mixed scenario when the qm_cluster
                            # is periodic in one direction
                            # (relevant for dislocation or crack cells)
                            # (qm cluster cell must be the same as
                            # the original cell in periodic direction
                            # and DIFFERENT form the original cell
                            # in non periodic directions
                            # three tests for three different directions
                            {"test_size": [4, 4, 1],
                             "expected_pbc": np.array([False, False, True]),
                             "buffer_width": 3.61},
                            {"test_size": [4, 1, 4],
                             "expected_pbc": np.array([False, True, False]),
                             "buffer_width": 3.61},
                            {"test_size": [1, 4, 4],
                             "expected_pbc": np.array([True, False, False]),
                             "buffer_width": 3.61},
                             # testing scenario periodic in one direction
                             # and non periodic in the other two
                             # relevant for surfaces.
                             # testing three different scenarios
                             {"test_size": [1, 1, 4],
                              "expected_pbc": np.array([True, True, False]),
                              "buffer_width": 3.61},
                             {"test_size": [4, 1, 1],
                              "expected_pbc": np.array([False, True, True]),
                              "buffer_width": 3.61},
                             {"test_size": [1, 4, 1],
                              "expected_pbc": np.array([True, False, True]),
                              "buffer_width": 3.61}
                         ])
def test_qm_pbc(kwargs, qm_calc, mm_calc, bulk_at):
    kwargs1 = {}
    kwargs1.update(kwargs)
    compare_qm_cell_and_pbc(qm_calc, mm_calc, bulk_at, **kwargs1)


def test_rescaled_calculator():
    """
    Test rescaled RescaledCalculator() by computing lattice constant
    and bulk modulus using fit to equation of state
    and comparing it to the desired values
    """

    # A simple empirical N-body potential for
    # transition metals by M. W. Finnis & J.E. Sinclair
    # https://www.tandfonline.com/doi/abs/10.1080/01418618408244210
    # using analytical formulation in order to avoid extra file dependence
    # All the constants are taken from the paper.
    # Please refer to the paper for more details

    def pair_potential(r):
        """
        returns the pair potential as a equation 27 in pair_potential
        r - numpy array with the values of distance to compute the pair function
        """
        # parameters for W
        c = 3.25
        c0 = 47.1346499
        c1 = -33.7665655
        c2 = 6.2541999

        energy = (c0 + c1 * r + c2 * r ** 2.0) * (r - c) ** 2.0
        energy[r > c] = 0.0

        return energy

    def cohesive_potential(r):
        """
        returns the cohesive potential as a equation 28 in pair_potential
        r - numpy array with the values of distance to compute the pair function
        """
        # parameters for W
        d = 4.400224

        rho = (r - d) ** 2.0
        rho[r > d] = 0.0

        return rho

    def embedding_function(rho):
        """
        returns energy as a function of electronic density from eq 3
        """

        A = 1.896373
        energy = - A * np.sqrt(rho)

        return energy

    cutoff = 4.400224
    W_FS = EAM(elements=['W'], embedded_energy=np.array([embedding_function]),
               electron_density=np.array([[cohesive_potential]]),
               phi=np.array([[pair_potential]]), cutoff=cutoff, form='fs')

    # compute MM and QM equations of state
    def strain(at, e, calc):
        at = at.copy()
        at.set_cell((1.0 + e) * at.cell, scale_atoms=True)
        at.calc = calc
        v = at.get_volume()
        e = at.get_potential_energy()
        return v, e

    # desired DFT values
    a0_qm = 3.18556
    C11_qm = 522  # pm 15 GPa
    C12_qm = 193  # pm 5 GPa
    B_qm = (C11_qm + 2.0 * C12_qm) / 3.0

    bulk_at = bulk("W", cubic=True)

    mm_calc = W_FS
    eps = np.linspace(-0.01, 0.01, 13)
    v_mm, E_mm = zip(*[strain(bulk_at, e, mm_calc) for e in eps])

    eos_mm = EquationOfState(v_mm, E_mm)
    v0_mm, _E0_mm, B_mm = eos_mm.fit()
    B_mm /= GPa
    a0_mm = v0_mm ** (1.0 / 3.0)

    mm_r = RescaledCalculator(mm_calc, a0_qm, B_qm, a0_mm, B_mm)
    bulk_at = bulk("W", cubic=True, a=a0_qm)
    v_mm_r, E_mm_r = zip(*[strain(bulk_at, e, mm_r) for e in eps])

    eos_mm_r = EquationOfState(v_mm_r, E_mm_r)
    v0_mm_r, _E0_mm_r, B_mm_r = eos_mm_r.fit()
    B_mm_r /= GPa
    a0_mm_r = v0_mm_r ** (1.0 / 3)

    # check match of a0 and B after rescaling is adequate
    # 0.1% error in lattice constant and bulk modulus
    assert abs((a0_mm_r - a0_qm) / a0_qm) < 1e-3
    assert abs((B_mm_r - B_qm) / B_qm) < 1e-3


@pytest.mark.slow()
def test_forceqmmm(qm_calc, mm_calc, bulk_at):

    # parameters
    N_cell = 2
    R_QMs = np.array([3, 7])

    sigma = (bulk_at * 2).get_distance(0, 1) * (2. ** (-1. / 6))

    at0 = bulk_at * N_cell
    r = at0.get_distances(0, np.arange(1, len(at0)), mic=True)
    print(len(r))
    del at0[0]  # introduce a vacancy
    print("N_cell", N_cell, 'N_MM', len(at0),
          "Size", N_cell * bulk_at.cell[0, 0])

    ref_at = at0.copy()
    ref_at.calc = qm_calc
    opt = FIRE(ref_at)
    opt.run(fmax=1e-3)
    u_ref = ref_at.positions - at0.positions

    us = []
    for R_QM in R_QMs:
        at = at0.copy()
        qm_mask = r < R_QM
        qm_buffer_mask_ref = r < 2 * qm_calc.rc + R_QM
        print(f'R_QM             {R_QM}   N_QM        {qm_mask.sum()}')
        print(f'R_QM + buffer: {2 * qm_calc.rc + R_QM:.2f}'
              f' N_QM_buffer {qm_buffer_mask_ref.sum()}')
        print(f'                     N_total:    {len(at)}')
        # Warning: Small size of the cell and large size of the buffer
        # lead to the qm calculation performed on the whole cell.
        qmmm = ForceQMMM(at, qm_mask, qm_calc, mm_calc,
                         buffer_width=2 * qm_calc.rc)
        qmmm.initialize_qm_buffer_mask(at)
        at.calc = qmmm
        opt = FIRE(at)
        opt.run(fmax=1e-3)
        us.append(at.positions - at0.positions)

    # compute error in energy norm |\nabla u - \nabla u_ref|
    def strain_error(at0, u_ref, u, cutoff, mask):
        I, J = neighbor_list('ij', at0, cutoff)
        I, J = np.array([(i, j) for i, j in zip(I, J) if mask[i]]).T
        v = u_ref - u
        dv = np.linalg.norm(v[I, :] - v[J, :], axis=1)
        return np.linalg.norm(dv)

    du_global = [strain_error(at0, u_ref, u, 1.5 * sigma,
                              np.ones(len(r))) for u in us]
    du_local = [strain_error(at0, u_ref, u, 1.5 * sigma, r < 3.0) for u in us]

    print('du_local', du_local)
    print('du_global', du_global)

    # check local errors are monotonically decreasing
    assert np.all(np.diff(du_local) < 0)

    # check global errors are monotonically converging
    assert np.all(np.diff(du_global) < 0)

    # biggest QM/MM should match QM result
    assert du_local[-1] < 1e-10
    assert du_global[-1] < 1e-10


@pytest.fixture()
def at0(qm_calc, mm_calc, bulk_at):
    alat = bulk_at.cell[0, 0]
    at0 = bulk_at * 5
    r = at0.get_distances(0, np.arange(len(at0)), mic=True)
    # should give 12 nearest neighbours + atom in the center
    R_QM = alat / np.sqrt(2.0) + 1.0e-3
    qm_mask = r < R_QM

    qmmm = ForceQMMM(at0, qm_mask, qm_calc, mm_calc,
                     buffer_width=3.61)

    qmmm.initialize_qm_buffer_mask(at0)
    at0.calc = qmmm

    return at0


def test_export_xyz(at0, testdir):
    """
    test the export_extxyz function and checks the region adn forces arrays
    """

    # evaluating forces to test exporting of forces
    forces = at0.get_forces()
    filename = "qmmm_export_test.xyz"

    qmmm = at0.calc
    qmmm.export_extxyz(filename=filename)

    from ase.io import read
    read_atoms = read(filename)

    assert "region" in read_atoms.arrays
    original_region = qmmm.get_region_from_masks()
    assert all(original_region == read_atoms.get_array("region"))

    assert "forces" in read_atoms.calc.results
    # absolute tolerance for comparing forces close to zero
    np.testing.assert_allclose(forces, read_atoms.get_forces(), atol=1.0e-6)


def test_set_masks_from_region(at0, qm_calc, mm_calc):
    """
    Test setting masks from region array
    """

    qmmm = at0.calc
    region = qmmm.get_region_from_masks(at0)

    # initialise another qmmm with different masks
    r = at0.get_distances(0, np.arange(len(at0)), mic=True)
    R_QM = 1.0e-3
    qm_mask = r < R_QM

    test_qmmm = ForceQMMM(at0, qm_mask, qm_calc, mm_calc,
                          buffer_width=3.61)

    # assert that number of qm atoms is different
    assert np.count_nonzero(qmmm.qm_selection_mask) != np.count_nonzero(
        test_qmmm.qm_selection_mask
    )

    test_qmmm.set_masks_from_region(region)

    assert all(test_qmmm.qm_selection_mask == qmmm.qm_selection_mask)
    assert all(test_qmmm.qm_buffer_mask == qmmm.qm_buffer_mask)

    test_region = test_qmmm.get_region_from_masks(at0)
    assert all(region == test_region)


def test_import_xyz(at0, qm_calc, mm_calc, testdir):
    """
    test the import_extxyz function and checks the mapping
    """

    filename = "qmmm_export_test.xyz"

    qmmm = at0.calc
    qmmm.export_extxyz(filename=filename, atoms=at0)

    imported_qmmm = ForceQMMM.import_extxyz(filename, qm_calc, mm_calc)

    assert all(imported_qmmm.qm_selection_mask == qmmm.qm_selection_mask)
    assert all(imported_qmmm.qm_buffer_mask == qmmm.qm_buffer_mask)



"""Test write and read."""
import io
import re

import numpy as np
import pytest

from ase import Atoms
from ase.build import bulk
from ase.data import atomic_numbers
from ase.io.lammpsdata import read_lammps_data, write_lammps_data


def compare(atoms: Atoms, atoms_ref: Atoms):
    """Compare two `Atoms` objects"""
    assert all(atoms.numbers == atoms_ref.numbers)
    assert atoms.get_masses() == pytest.approx(atoms_ref.get_masses())

    # Note: Raw positions cannot be compared.
    # `write_lammps_data` changes cell orientation.
    assert atoms.get_scaled_positions() == pytest.approx(
        atoms_ref.get_scaled_positions())


@pytest.mark.parametrize('masses', [False, True])
class _Base:
    def run(self, atoms_ref: Atoms, masses: bool):
        """Run tests"""
        self.check_explicit_numbers(atoms_ref, masses)
        if masses:
            self.check_masses2numbers(atoms_ref)

    def check_explicit_numbers(self, atoms_ref: Atoms, masses: bool):
        """Check if write-read is consistent when giving Z_of_type."""
        buf = io.StringIO()
        write_lammps_data(buf, atoms_ref, masses=masses)
        buf.seek(0)
        # By default, write_lammps_data assigns atom types to the elements in
        # alphabetical order. To be consistent, here spiecies are also sorted.
        species = sorted(set(atoms_ref.get_chemical_symbols()))
        Z_of_type = {i + 1: atomic_numbers[s] for i, s in enumerate(species)}
        atoms = read_lammps_data(buf, Z_of_type=Z_of_type, atom_style='atomic')
        compare(atoms, atoms_ref)

    def check_masses2numbers(self, atoms_ref: Atoms):
        """Check if write-read is consistent when guessing atomic numbers."""
        buf = io.StringIO()
        write_lammps_data(buf, atoms_ref, masses=True)
        buf.seek(0)
        atoms = read_lammps_data(buf, atom_style='atomic')
        compare(atoms, atoms_ref)


@pytest.mark.parametrize('cubic', [False, True])
class TestCubic(_Base):
    """Test cubic structures."""

    def test_bcc(self, cubic: bool, masses: bool):
        """Test bcc."""
        atoms_ref = bulk('Li', 'bcc', cubic=cubic)
        self.run(atoms_ref, masses)

    def test_fcc(self, cubic: bool, masses: bool):
        """Test fcc."""
        atoms_ref = bulk('Cu', 'fcc', cubic=cubic)
        self.run(atoms_ref, masses)

    def test_rocksalt(self, cubic: bool, masses: bool):
        """Test rocksalt."""
        atoms_ref = bulk('NaCl', 'rocksalt', a=1.0, cubic=cubic)
        self.run(atoms_ref, masses)

    def test_fluorite(self, cubic: bool, masses: bool):
        """Test fluorite."""
        atoms_ref = bulk('CaF2', 'fluorite', a=1.0, cubic=cubic)
        self.run(atoms_ref, masses)


@pytest.mark.parametrize('orthorhombic', [False, True])
class TestOrthorhombic(_Base):
    """Test orthorhombic structures."""

    def test_hcp(self, orthorhombic: bool, masses: bool):
        """Test hcp."""
        atoms_ref = bulk('Mg', 'hcp', orthorhombic=orthorhombic)
        self.run(atoms_ref, masses)


@pytest.mark.parametrize('atom_style', ['atomic', 'charge', 'full'])
def test_atom_style(atom_style: str):
    """Test `atom_style`"""
    atoms_ref = bulk('Cu', 'fcc')

    # note that we should do `masses=True` to later guess atomic numbers
    buf = io.StringIO()
    write_lammps_data(buf, atoms_ref, masses=True, atom_style=atom_style)

    # test if the `atom_style` can be guessed from the comment
    buf.seek(0)
    atoms = read_lammps_data(buf, atom_style=None)
    compare(atoms, atoms_ref)

    # test when `atom_style` is explicitly specified
    buf.seek(0)
    atoms = read_lammps_data(buf, atom_style=atom_style)
    compare(atoms, atoms_ref)

    if atom_style not in ['atomic', 'full']:
        return

    # test if `atom_style` can be guessed from the length of fields
    # remove comment in the 'Atoms' line
    buf = io.StringIO(re.sub('.*Atoms.*#.*', 'Atoms', buf.getvalue()))
    atoms = read_lammps_data(buf, atom_style=None)
    compare(atoms, atoms_ref)


@pytest.mark.parametrize('atom_style', ['atomic', 'charge', 'full'])
@pytest.mark.parametrize('write_image_flags', [False, True])
def test_image_flags(write_image_flags: bool, atom_style: str):
    """Test if `wrap` and `write_image_flags` work correctly."""
    atoms_ref = bulk('Ge')

    # shift atomic positions
    scaled_positions = atoms_ref.get_scaled_positions(wrap=False)
    shift = (0.125, 1.125, -0.125)
    atoms_ref.set_scaled_positions(scaled_positions + shift)

    # note that we should do `masses=True` to later guess atomic numbers
    buf = io.StringIO()
    write_lammps_data(
        buf,
        atoms_ref,
        write_image_flags=write_image_flags,
        masses=True,
        atom_style=atom_style,
    )

    buf.seek(0)
    atoms = read_lammps_data(buf)

    # The atomic positions should be wrapped.
    np.testing.assert_allclose(
        atoms.get_scaled_positions(wrap=False),
        atoms_ref.get_scaled_positions(wrap=False),
    )


def test_bonds(lammpsdata_file_path):
    """Test if writing bonds works correctly."""
    atoms = read_lammps_data(lammpsdata_file_path, atom_style='full')
    lammpsdata_buf = io.StringIO()
    write_lammps_data(
        lammpsdata_buf, atoms, atom_style='full',
        masses=True, velocities=True, bonds=True)
    lammpsdata_buf.seek(0)
    atoms2 = read_lammps_data(lammpsdata_buf, atom_style='full')
    np.testing.assert_array_equal(
        atoms.arrays["bonds"], atoms2.arrays["bonds"]
    )



raise ImportError(
    'The ase.ga.bulk_startgenerator module has been deprecated. '
    'The same functionality is now provided by the '
    'ase.ga.startgenerator module. Please consult its documentation '
    'to verify how to e.g. initialize a StartGenerator object.')



import pytest

from ase.atoms import Atoms


@pytest.mark.calculator_lite()
@pytest.mark.calculator('lammpsrun')
def test_no_data_file_wrap(factory):
    """
    If 'create_atoms' hasn't been given the appropriate 'remap yes' option,
    atoms falling outside of a periodic cell are not actually created.  The
    lammpsrun calculator will then look at the thermo output and determine a
    discrepancy the number of atoms reported compared to the length of the
    ASE Atoms object and raise a RuntimeError.  This problem can only
    possibly arise when the 'no_data_file' option for the calculator is set
    to True.  Furthermore, note that if atoms fall outside of the box along
    non-periodic dimensions, create_atoms is going to refuse to create them
    no matter what, so you simply can't use the 'no_data_file' option if you
    want to allow for that scenario.
    """

    # Make a periodic box and put one atom outside of it
    pos = [[0.0, 0.0, 0.0], [-2.0, 0.0, 0.0]]
    atoms = Atoms(symbols=["Ar"] * 2, positions=pos, cell=[10.0, 10.0, 10.0],
                  pbc=True)

    # Set parameters for calculator
    params = {}
    params["pair_style"] = "lj/cut 8.0"
    params["pair_coeff"] = ["1 1 0.0108102 3.345"]

    # Don't write a data file string. This will force
    # ase.calculators.lammps.inputwriter.write_lammps_in to write a bunch of
    # 'create_atoms' commands into the LAMMPS input file
    params["no_data_file"] = True

    with factory.calc(specorder=["Ar"], **params) as calc:
        atoms.calc = calc
        atoms.get_potential_energy()
        # assert something?



import numpy as np

from ase import Atoms
from ase.calculators.calculator import compare_atoms


def test_compare_atoms():
    """
    Check that Atoms.compare_atoms correctly accounts for the different
    types of system changes
    """

    # A system property that's an attribute of Atoms, but isn't in
    # Atoms.arrays (currently this is just 'cell' and 'pbc')
    cell1 = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
    cell2 = cell1 * 2
    atoms1 = Atoms(cell=cell1)
    atoms2 = Atoms(cell=cell2)
    assert set(compare_atoms(atoms1, atoms2)) == {"cell"}

    # A system property other than 'initial_charges' or 'initial_magmoms'
    # that exists in the `arrays` attribute of one Atoms object but not the
    # other
    atoms1 = Atoms()
    atoms2 = Atoms(numbers=[0], positions=[[0, 0, 0]])
    assert set(compare_atoms(atoms1, atoms2)) == {"positions", "numbers"}

    # A change in a system property that exists in the `arrays` attribute
    # of both Atoms objects passed into this function
    atoms1 = Atoms(numbers=[0], positions=[[0, 0, 0]])
    atoms2 = Atoms(numbers=[0], positions=[[1, 0, 0]])
    assert set(compare_atoms(atoms1, atoms2)) == {"positions"}

    # An excluded property (re-use atoms1 and atoms2 from previous check)
    assert set(compare_atoms(atoms1, atoms2,
               excluded_properties={"positions"})) == set()

    # Optional array (currently 'initial_charges' or 'initial_magmoms')
    # NOTE: Suppose you initialize an array of *zero charges* for atoms2
    #       but not atoms1.  The current compare_atoms function will still
    #       indicate that 'initial_charges' is in system_changes simply
    #       because it isn't in both of them.  This is despite the fact that
    #       if one were to call atoms1.get_initial_charges, you would get
    #       back an array of zeros.  However, this scenario should only ever
    #       occur rarely.
    atoms1 = Atoms(numbers=[0], positions=[[0, 0, 0]])
    atoms2 = Atoms(numbers=[0], positions=[[0, 0, 0]], charges=[1.13])
    assert set(compare_atoms(atoms1, atoms2)) == {"initial_charges"}



"""Methods for finding minimum-energy paths and/or saddle points."""

from ase.mep.autoneb import AutoNEB
from ase.mep.dimer import DimerControl, MinModeAtoms, MinModeTranslate
from ase.mep.dyneb import DyNEB
from ase.mep.neb import (
    NEB,
    NEBTools,
    SingleCalculatorNEB,
    idpp_interpolate,
    interpolate,
)

__all__ = ['NEB', 'NEBTools', 'DyNEB', 'AutoNEB', 'interpolate',
           'idpp_interpolate', 'SingleCalculatorNEB',
           'DimerControl', 'MinModeAtoms', 'MinModeTranslate']



from ase import Atoms


def test_bandpath_kpts_axis():
    # See https://gitlab.com/ase/ase/issues/502

    a = 3.16
    atoms = Atoms(cell=[a, a, 12, 90, 90, 120], pbc=True)
    G = [0, 0, 0]
    K = [1 / 3., 1 / 3., 0]
    K_ = [-1 / 3., -1 / 3., 0]
    path = [K, G, K_]
    nspecial_points = len(path)

    for npoints in [10, 11]:
        bandpath = atoms.cell.bandpath(path, npoints=npoints)
        kpts, x, X = bandpath.get_linear_kpoint_axis()
        assert len(kpts) == npoints
        assert len(x) == nspecial_points
        assert len(X) == nspecial_points



"""ASE calculator for the LAMMPS classical MD code"""
# lammps.py (2011/03/29)
#
# Copyright (C) 2009 - 2011 Joerg Meyer, joerg.meyer@ch.tum.de
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this file; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA or see <http://www.gnu.org/licenses/>.


import os
import shlex
import shutil
import subprocess
import warnings
from re import IGNORECASE
from re import compile as re_compile
from tempfile import NamedTemporaryFile, mkdtemp
from tempfile import mktemp as uns_mktemp
from threading import Thread
from typing import Any, Dict

import numpy as np

from ase.calculators.calculator import Calculator, all_changes
from ase.calculators.lammps import (
    CALCULATION_END_MARK,
    Prism,
    convert,
    write_lammps_in,
)
from ase.data import atomic_masses, chemical_symbols
from ase.io.lammpsdata import write_lammps_data
from ase.io.lammpsrun import read_lammps_dump

__all__ = ["LAMMPS"]


class LAMMPS(Calculator):
    """LAMMPS (https://lammps.sandia.gov/) calculator

    The environment variable :envvar:`ASE_LAMMPSRUN_COMMAND` must be defined to
    tell ASE how to call a LAMMPS binary. This should contain the path to the
    lammps binary, or more generally, a command line possibly also including an
    MPI-launcher command.

    For example (in a Bourne-shell compatible environment):

    .. code-block:: bash

        export ASE_LAMMPSRUN_COMMAND=/path/to/lmp_binary

    or possibly something similar to

    .. code-block:: bash

        export ASE_LAMMPSRUN_COMMAND="/path/to/mpirun --np 4 lmp_binary"

    Parameters
    ----------
    files : list[str]
        List of files needed by LAMMPS. Typically a list of potential files.
    parameters : dict[str, Any]
        Dictionary of settings to be passed into the input file for calculation.
    specorder : list[str]
        Within LAAMPS, atoms are identified by an integer value starting from 1.
        This variable allows the user to define the order of the indices
        assigned to the atoms in the calculation, with the default
        if not given being alphabetical
    keep_tmp_files : bool, default: False
        Retain any temporary files created. Mostly useful for debugging.
    tmp_dir : str, default: None
        path/dirname (default None -> create automatically).
        Explicitly control where the calculator object should create
        its files. Using this option implies 'keep_tmp_files=True'.
    no_data_file : bool, default: False
        Controls whether an explicit data file will be used for feeding
        atom coordinates into lammps. Enable it to lessen the pressure on
        the (tmp) file system. THIS OPTION MIGHT BE UNRELIABLE FOR CERTAIN
        CORNER CASES (however, if it fails, you will notice...).
    keep_alive : bool, default: True
        When using LAMMPS as a spawned subprocess, keep the subprocess
        alive (but idling when unused) along with the calculator object.
    always_triclinic : bool, default: False
        Force LAMMPS to treat the cell as tilted, even if the cell is not
        tilted, by printing ``xy``, ``xz``, ``yz`` in the data file.
    reduce_cell : bool, default: False
        If True, reduce cell to have shorter lattice vectors.
    write_velocities : bool, default: False
        If True, forward ASE velocities to LAMMPS.
    verbose: bool, default: False
        If True, print additional debugging output to STDOUT.

    Examples
    --------
    Provided that the respective potential file is in the working directory,
    one can simply run (note that LAMMPS needs to be compiled to work with EAM
    potentials)

    ::

        from ase import Atom, Atoms
        from ase.build import bulk
        from ase.calculators.lammpsrun import LAMMPS

        parameters = {'pair_style': 'eam/alloy',
                    'pair_coeff': ['* * NiAlH_jea.eam.alloy H Ni']}

        files = ['NiAlH_jea.eam.alloy']

        Ni = bulk('Ni', cubic=True)
        H = Atom('H', position=Ni.cell.diagonal()/2)
        NiH = Ni + H

        lammps = LAMMPS(files=files, **parameters)

        NiH.calc = lammps
        print("Energy ", NiH.get_potential_energy())
    """

    name = "lammpsrun"
    implemented_properties = ["energy", "free_energy", "forces", "stress",
                              "energies"]

    # parameters to choose options in LAMMPSRUN
    ase_parameters: Dict[str, Any] = dict(
        specorder=None,
        atorder=True,
        always_triclinic=False,
        reduce_cell=False,
        keep_alive=True,
        keep_tmp_files=False,
        no_data_file=False,
        tmp_dir=None,
        files=[],  # usually contains potential parameters
        verbose=False,
        write_velocities=False,
        binary_dump=True,  # bool - use binary dump files (full
                           # precision but long long ids are casted to
                           # double)
        lammps_options="-echo log -screen none -log /dev/stdout",
        trajectory_out=None,  # file object, if is not None the
                              # trajectory will be saved in it
    )

    # parameters forwarded to LAMMPS
    lammps_parameters = dict(
        boundary=None,  # bounadry conditions styles
        units="metal",  # str - Which units used; some potentials
                        # require certain units
        atom_style="atomic",
        special_bonds=None,
        # potential informations
        pair_style="lj/cut 2.5",
        pair_coeff=["* * 1 1"],
        masses=None,
        pair_modify=None,
        # variables controlling the output
        thermo_args=[
            "step", "temp", "press", "cpu", "pxx", "pyy", "pzz",
            "pxy", "pxz", "pyz", "ke", "pe", "etotal", "vol", "lx",
            "ly", "lz", "atoms", ],
        dump_properties=["id", "type", "x", "y", "z", "vx", "vy",
                         "vz", "fx", "fy", "fz", ],
        dump_period=1,  # period of system snapshot saving (in MD steps)
    )

    default_parameters = dict(ase_parameters, **lammps_parameters)

    def __init__(self, label="lammps", **kwargs):
        super().__init__(label=label, **kwargs)

        self.prism = None
        self.calls = 0
        self.forces = None
        # thermo_content contains data "written by" thermo_style.
        # It is a list of dictionaries, each dict (one for each line
        # printed by thermo_style) contains a mapping between each
        # custom_thermo_args-argument and the corresponding
        # value as printed by lammps. thermo_content will be
        # re-populated by the read_log method.
        self.thermo_content = []

        if self.parameters['tmp_dir'] is not None:
            # If tmp_dir is pointing somewhere, don't remove stuff!
            self.parameters['keep_tmp_files'] = True
        self._lmp_handle = None  # To handle the lmp process

        if self.parameters['tmp_dir'] is None:
            self.parameters['tmp_dir'] = mkdtemp(prefix="LAMMPS-")
        else:
            self.parameters['tmp_dir'] = os.path.realpath(
                self.parameters['tmp_dir'])
            if not os.path.isdir(self.parameters['tmp_dir']):
                os.mkdir(self.parameters['tmp_dir'], 0o755)

        for f in self.parameters['files']:
            shutil.copy(
                f, os.path.join(self.parameters['tmp_dir'],
                                os.path.basename(f)))

    def get_lammps_command(self):
        cmd = self.parameters.get('command')

        if cmd is None:
            from ase.config import cfg
            envvar = f'ASE_{self.name.upper()}_COMMAND'
            cmd = cfg.get(envvar)

        if cmd is None:
            # TODO deprecate and remove guesswork
            cmd = 'lammps'

        opts = self.parameters.get('lammps_options')

        if opts is not None:
            cmd = f'{cmd} {opts}'

        return cmd

    def clean(self, force=False):

        self._lmp_end()

        if not self.parameters['keep_tmp_files'] or force:
            shutil.rmtree(self.parameters['tmp_dir'])

    def check_state(self, atoms, tol=1.0e-10):
        # Transforming the unit cell to conform to LAMMPS' convention for
        # orientation (c.f. https://lammps.sandia.gov/doc/Howto_triclinic.html)
        # results in some precision loss, so we use bit larger tolerance than
        # machine precision here.  Note that there can also be precision loss
        # related to how many significant digits are specified for things in
        # the LAMMPS input file.
        return Calculator.check_state(self, atoms, tol)

    def calculate(self, atoms=None, properties=None, system_changes=None):
        if properties is None:
            properties = self.implemented_properties
        if system_changes is None:
            system_changes = all_changes
        Calculator.calculate(self, atoms, properties, system_changes)
        self.run()

    def _lmp_alive(self):
        # Return True if this calculator is currently handling a running
        # lammps process
        return self._lmp_handle and not isinstance(
            self._lmp_handle.poll(), int
        )

    def _lmp_end(self):
        # Close lammps input and wait for lammps to end. Return process
        # return value
        if self._lmp_alive():
            # !TODO: handle lammps error codes
            try:
                self._lmp_handle.communicate(timeout=5)
            except subprocess.TimeoutExpired:
                self._lmp_handle.kill()
                self._lmp_handle.communicate()
            err = self._lmp_handle.poll()
            assert err is not None
            return err

    def set_missing_parameters(self):
        """Verify that all necessary variables are set.
        """
        symbols = self.atoms.get_chemical_symbols()
        # If unspecified default to atom types in alphabetic order
        if not self.parameters.get('specorder'):
            self.parameters['specorder'] = sorted(set(symbols))

        # !TODO: handle cases were setting masses actual lead to errors
        if not self.parameters.get('masses'):
            self.parameters['masses'] = []
            for type_id, specie in enumerate(self.parameters['specorder']):
                mass = atomic_masses[chemical_symbols.index(specie)]
                self.parameters['masses'] += [
                    f"{type_id + 1:d} {mass:f}"
                ]

        # set boundary condtions
        if not self.parameters.get('boundary'):
            b_str = " ".join(["fp"[int(x)] for x in self.atoms.pbc])
            self.parameters['boundary'] = b_str

    def run(self, set_atoms=False):
        # !TODO: split this function
        """Method which explicitly runs LAMMPS."""
        pbc = self.atoms.get_pbc()
        if all(pbc):
            cell = self.atoms.get_cell()
        elif not any(pbc):
            # large enough cell for non-periodic calculation -
            # LAMMPS shrink-wraps automatically via input command
            #       "periodic s s s"
            # below
            cell = 2 * np.max(np.abs(self.atoms.get_positions())) * np.eye(3)
        else:
            warnings.warn(
                "semi-periodic ASE cell detected - translation "
                + "to proper LAMMPS input cell might fail"
            )
            cell = self.atoms.get_cell()
        self.prism = Prism(cell)

        self.set_missing_parameters()
        self.calls += 1

        # change into subdirectory for LAMMPS calculations
        tempdir = self.parameters['tmp_dir']

        # setup file names for LAMMPS calculation
        label = f"{self.label}{self.calls:>06}"
        lammps_in = uns_mktemp(
            prefix="in_" + label, dir=tempdir
        )
        lammps_log = uns_mktemp(
            prefix="log_" + label, dir=tempdir
        )
        lammps_trj_fd = NamedTemporaryFile(
            prefix="trj_" + label,
            suffix=(".bin" if self.parameters['binary_dump'] else ""),
            dir=tempdir,
            delete=(not self.parameters['keep_tmp_files']),
        )
        lammps_trj = lammps_trj_fd.name
        if self.parameters['no_data_file']:
            lammps_data = None
        else:
            lammps_data_fd = NamedTemporaryFile(
                prefix="data_" + label,
                dir=tempdir,
                delete=(not self.parameters['keep_tmp_files']),
                mode='w',
                encoding='ascii'
            )
            write_lammps_data(
                lammps_data_fd,
                self.atoms,
                specorder=self.parameters['specorder'],
                force_skew=self.parameters['always_triclinic'],
                reduce_cell=self.parameters['reduce_cell'],
                velocities=self.parameters['write_velocities'],
                prismobj=self.prism,
                units=self.parameters['units'],
                atom_style=self.parameters['atom_style'],
            )
            lammps_data = lammps_data_fd.name
            lammps_data_fd.flush()

        # see to it that LAMMPS is started
        if not self._lmp_alive():
            command = self.get_lammps_command()
            # Attempt to (re)start lammps
            self._lmp_handle = subprocess.Popen(
                shlex.split(command, posix=(os.name == "posix")),
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                encoding='ascii',
            )
        lmp_handle = self._lmp_handle

        # Create thread reading lammps stdout (for reference, if requested,
        # also create lammps_log, although it is never used)
        if self.parameters['keep_tmp_files']:
            lammps_log_fd = open(lammps_log, "w")
            fd = SpecialTee(lmp_handle.stdout, lammps_log_fd)
        else:
            fd = lmp_handle.stdout
        thr_read_log = Thread(target=self.read_lammps_log, args=(fd,))
        thr_read_log.start()

        # write LAMMPS input (for reference, also create the file lammps_in,
        # although it is never used)
        if self.parameters['keep_tmp_files']:
            lammps_in_fd = open(lammps_in, "w")
            fd = SpecialTee(lmp_handle.stdin, lammps_in_fd)
        else:
            fd = lmp_handle.stdin
        write_lammps_in(
            lammps_in=fd,
            parameters=self.parameters,
            atoms=self.atoms,
            prismobj=self.prism,
            lammps_trj=lammps_trj,
            lammps_data=lammps_data,
        )

        if self.parameters['keep_tmp_files']:
            lammps_in_fd.close()

        # Wait for log output to be read (i.e., for LAMMPS to finish)
        # and close the log file if there is one
        thr_read_log.join()
        if self.parameters['keep_tmp_files']:
            lammps_log_fd.close()

        if not self.parameters['keep_alive']:
            self._lmp_end()

        exitcode = lmp_handle.poll()
        if exitcode and exitcode != 0:
            raise RuntimeError(
                "LAMMPS exited in {} with exit code: {}."
                "".format(tempdir, exitcode)
            )

        # A few sanity checks
        if len(self.thermo_content) == 0:
            raise RuntimeError("Failed to retrieve any thermo_style-output")
        if int(self.thermo_content[-1]["atoms"]) != len(self.atoms):
            # This obviously shouldn't happen, but if prism.fold_...() fails,
            # it could
            raise RuntimeError("Atoms have gone missing")

        trj_atoms = read_lammps_dump(
            infileobj=lammps_trj,
            order=self.parameters['atorder'],
            index=-1,
            prismobj=self.prism,
            specorder=self.parameters['specorder'],
        )

        if set_atoms:
            self.atoms = trj_atoms.copy()

        self.forces = trj_atoms.get_forces()
        # !TODO: trj_atoms is only the last snapshot of the system; Is it
        #        desirable to save also the inbetween steps?
        if self.parameters['trajectory_out'] is not None:
            # !TODO: is it advisable to create here temporary atoms-objects
            self.trajectory_out.write(trj_atoms)

        tc = self.thermo_content[-1]
        self.results["energy"] = convert(
            tc["pe"], "energy", self.parameters["units"], "ASE"
        )
        self.results["free_energy"] = self.results["energy"]
        self.results['forces'] = convert(self.forces.copy(),
                                         'force',
                                         self.parameters['units'],
                                         'ASE')
        stress = np.array(
            [-tc[i] for i in ("pxx", "pyy", "pzz", "pyz", "pxz", "pxy")]
        )

        # We need to apply the Lammps rotation stuff to the stress:
        xx, yy, zz, yz, xz, xy = stress
        stress_tensor = np.array([[xx, xy, xz],
                                  [xy, yy, yz],
                                  [xz, yz, zz]])
        stress_atoms = self.prism.tensor2_to_ase(stress_tensor)
        stress_atoms = stress_atoms[[0, 1, 2, 1, 0, 0],
                                    [0, 1, 2, 2, 2, 1]]
        stress = stress_atoms

        self.results["stress"] = convert(
            stress, "pressure", self.parameters["units"], "ASE"
        )

        lammps_trj_fd.close()
        if not self.parameters['no_data_file']:
            lammps_data_fd.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self._lmp_end()

    def read_lammps_log(self, fileobj):
        # !TODO: somehow communicate 'thermo_content' explicitly
        """Method which reads a LAMMPS output log file."""

        # read_log depends on that the first (three) thermo_style custom args
        # can be capitalized and matched against the log output. I.e.
        # don't use e.g. 'ke' or 'cpu' which are labeled KinEng and CPU.
        mark_re = r"^\s*" + r"\s+".join(
            [x.capitalize() for x in self.parameters['thermo_args'][0:3]]
        )
        _custom_thermo_mark = re_compile(mark_re)

        # !TODO: regex-magic necessary?
        # Match something which can be converted to a float
        f_re = r"([+-]?(?:(?:\d+(?:\.\d*)?|\.\d+)(?:e[+-]?\d+)?|nan|inf))"
        n_args = len(self.parameters["thermo_args"])
        # Create a re matching exactly N white space separated floatish things
        _custom_thermo_re = re_compile(
            r"^\s*" + r"\s+".join([f_re] * n_args) + r"\s*$", flags=IGNORECASE
        )

        thermo_content = []
        line = fileobj.readline()
        while line and line.strip() != CALCULATION_END_MARK:
            # check error
            if 'ERROR:' in line:
                raise RuntimeError(f'LAMMPS exits with error message: {line}')

            # get thermo output
            if _custom_thermo_mark.match(line):
                while True:
                    line = fileobj.readline()
                    if 'WARNING:' in line:
                        continue

                    bool_match = _custom_thermo_re.match(line)
                    if not bool_match:
                        break

                    # create a dictionary between each of the
                    # thermo_style args and it's corresponding value
                    thermo_content.append(
                        dict(
                            zip(
                                self.parameters['thermo_args'],
                                map(float, bool_match.groups()),
                            )
                        )
                    )
            else:
                line = fileobj.readline()

        self.thermo_content = thermo_content


class SpecialTee:
    """A special purpose, with limited applicability, tee-like thing.

    A subset of stuff read from, or written to, orig_fd,
    is also written to out_fd.
    It is used by the lammps calculator for creating file-logs of stuff
    read from, or written to, stdin and stdout, respectively.
    """

    def __init__(self, orig_fd, out_fd):
        self._orig_fd = orig_fd
        self._out_fd = out_fd
        self.name = orig_fd.name

    def write(self, data):
        self._orig_fd.write(data)
        self._out_fd.write(data)
        self.flush()

    def read(self, *args, **kwargs):
        data = self._orig_fd.read(*args, **kwargs)
        self._out_fd.write(data)
        return data

    def readline(self, *args, **kwargs):
        data = self._orig_fd.readline(*args, **kwargs)
        self._out_fd.write(data)
        return data

    def readlines(self, *args, **kwargs):
        data = self._orig_fd.readlines(*args, **kwargs)
        self._out_fd.write("".join(data))
        return data

    def flush(self):
        self._orig_fd.flush()
        self._out_fd.flush()



import pytest
from numpy.testing import assert_allclose

from ase.data import atomic_masses, atomic_numbers
from ase.filters import UnitCellFilter
from ase.optimize import QuasiNewton
from ase.spacegroup import crystal


@pytest.mark.calculator_lite()
@pytest.mark.calculator('lammpsrun')
def test_NaCl_minimize(factory):
    a = 6.15
    n = 4
    nacl = crystal(['Na', 'Cl'], [(0, 0, 0), (0.5, 0.5, 0.5)], spacegroup=225,
                   cellpar=[a, a, a, 90, 90, 90]).repeat((n, n, n))

    # Buckingham parameters from
    # https://physics.stackexchange.com/questions/250018

    pair_style = 'buck/coul/long 12.0'
    pair_coeff = ['1 1 3796.9 0.2603 124.90']
    pair_coeff += ['2 2 1227.2 0.3214 124.90']
    pair_coeff += ['1 2 4117.9 0.3048 0.0']
    masses = ['1 {}'.format(atomic_masses[atomic_numbers['Na']]),
              '2 {}'.format(atomic_masses[atomic_numbers['Cl']])]

    with factory.calc(
            specorder=['Na', 'Cl'],
            pair_style=pair_style,
            pair_coeff=pair_coeff,
            masses=masses,
            atom_style='charge',
            kspace_style='pppm 1.0e-5',
            keep_tmp_files=True,
    ) as calc:

        for a in nacl:
            if a.symbol == 'Na':
                a.charge = +1.
            else:
                a.charge = -1.

        nacl.calc = calc

        assert_allclose(nacl.get_potential_energy(), -1896.216737561538,
                        atol=1e-4, rtol=1e-4)

        nacl.get_potential_energy()

        ucf = UnitCellFilter(nacl)
        with QuasiNewton(ucf) as dyn:
            dyn.run(fmax=1.0E-2)

        assert_allclose(nacl.get_potential_energy(), -1897.208861729178,
                        atol=1e-4, rtol=1e-4)



import numpy as np

import ase.units as units
from ase import Atoms
from ase.calculators.acn import (
    ACN,
    epsilon_c,
    epsilon_me,
    epsilon_n,
    m_me,
    r_cn,
    r_mec,
    sigma_c,
    sigma_me,
    sigma_n,
)
from ase.calculators.qmmm import EIQMMM, LJInteractionsGeneral, SimpleQMMM
from ase.constraints import FixLinearTriatomic
from ase.md.verlet import VelocityVerlet


def test_rattle_linear(testdir):
    """Test RATTLE and QM/MM for rigid linear acetonitrile."""

    sigma = np.array([sigma_me, sigma_c, sigma_n])
    epsilon = np.array([epsilon_me, epsilon_c, epsilon_n])
    i = LJInteractionsGeneral(sigma, epsilon, sigma, epsilon, 3)

    for calc in [ACN(),
                 SimpleQMMM([0, 1, 2], ACN(), ACN(), ACN()),
                 EIQMMM([0, 1, 2], ACN(), ACN(), i)]:

        dimer = Atoms('CCNCCN',
                      [(-r_mec, 0, 0),
                       (0, 0, 0),
                       (r_cn, 0, 0),
                       (r_mec, 3.7, 0),
                       (0, 3.7, 0),
                       (-r_cn, 3.7, 0)])

        masses = dimer.get_masses()
        masses[::3] = m_me
        dimer.set_masses(masses)

        fixd = FixLinearTriatomic(triples=[(0, 1, 2), (3, 4, 5)])

        dimer.set_constraint(fixd)

        dimer.calc = calc

        d1 = dimer[:3].get_all_distances()
        d2 = dimer[3:].get_all_distances()
        e = dimer.get_potential_energy()

        with VelocityVerlet(dimer, 2.0 * units.fs,
                            trajectory=calc.name + '.traj',
                            logfile=calc.name + '.log',
                            loginterval=20) as md:
            md.run(100)

        de = dimer.get_potential_energy() - e

        assert np.all(abs(dimer[:3].get_all_distances() - d1) < 1e-10)
        assert np.all(abs(dimer[3:].get_all_distances() - d2) < 1e-10)
        assert abs(de - -0.005) < 0.001



import numpy as np

from ase.build import bulk
from ase.calculators.harmonic import SpringCalculator
from ase.calculators.test import gradient_test


def test_springcalc():
    # setup
    k = 3.0
    atoms_ideal = bulk('Al').repeat(3)
    calc = SpringCalculator(atoms_ideal.get_positions(), k)
    displacements = np.array([(d, 2 * d, 3 * d)
                             for d in np.linspace(0, 1, len(atoms_ideal))])

    # calc forces and energy
    atoms = atoms_ideal.copy()
    atoms.positions += displacements
    atoms.calc = calc
    forces = atoms.get_forces()
    Epot = atoms.get_potential_energy()

    # reference forces and energy
    Epot_target = np.sum(k / 2.0 * displacements**2)
    forces_target = - k * displacements

    assert np.allclose(forces, forces_target)
    assert np.isclose(Epot, Epot_target)

    # numeric forces test
    atoms_ideal.calc = calc
    f, fn = gradient_test(atoms_ideal)
    assert abs(f - fn).max() < 1e-10



import numpy as np
from scipy.linalg import qr


def random_orthogonal_matrix(dim, rng=np.random, real=False):
    """Generate uniformly distributed random orthogonal matrices"""
    if real:
        from scipy.stats import special_ortho_group
        ortho_m = special_ortho_group.rvs(dim=dim, random_state=rng)
    else:
        # The best method but not supported on old systems
        # from scipy.stats import unitary_group
        # ortho_m = unitary_group.rvs(dim=dim, random_state=rng)

        # Alternative method from https://stackoverflow.com/questions/38426349
        H = rng.random((dim, dim))
        Q, R = qr(H)
        ortho_m = Q @ np.diag(np.sign(np.diag(R)))

    return ortho_m


def _empty():
    return np.empty(0, complex)


class WannierSpec:
    def __init__(self, Nk, Nw, Nb, fixedstates_k):
        self.Nk = Nk
        self.Nw = Nw
        self.Nb = Nb
        self.fixedstates_k = fixedstates_k

    def _zeros(self):
        return np.zeros((self.Nk, self.Nw, self.Nw), complex)

    def bloch(self, edf_k):
        U_kww = self._zeros()
        C_kul = []
        for U, M, L in zip(U_kww, self.fixedstates_k, edf_k):
            U[:] = np.identity(self.Nw, complex)
            if L > 0:
                C_kul.append(np.identity(self.Nb - M, complex)[:, :L])
            else:
                C_kul.append(_empty())
        return WannierState(C_kul, U_kww)

    def random(self, rng, edf_k):
        # Set U and C to random (orthogonal) matrices
        U_kww = self._zeros()
        C_kul = []
        for U, M, L in zip(U_kww, self.fixedstates_k, edf_k):
            U[:] = random_orthogonal_matrix(self.Nw, rng, real=False)
            if L > 0:
                C_kul.append(random_orthogonal_matrix(
                    self.Nb - M, rng=rng, real=False)[:, :L])
            else:
                C_kul.append(_empty())
        return WannierState(C_kul, U_kww)

    def initial_orbitals(self, calc, orbitals, kptgrid, edf_k, spin):
        C_kul, U_kww = calc.initial_wannier(
            orbitals, kptgrid, self.fixedstates_k, edf_k, spin, self.Nb)
        return WannierState(C_kul, U_kww)

    def initial_wannier(self, calc, method, kptgrid, edf_k, spin):
        C_kul, U_kww = calc.initial_wannier(
            method, kptgrid, self.fixedstates_k,
            edf_k, spin, self.Nb)
        return WannierState(C_kul, U_kww)

    def scdm(self, calc, kpt_kc, spin):
        from ase.dft.wannier import scdm

        # get the size of the grid and check if there are Nw bands:
        ps = calc.get_pseudo_wave_function(band=self.Nw,
                                           kpt=0, spin=0)
        Ng = ps.size
        pseudo_nkG = np.zeros((self.Nb, self.Nk, Ng),
                              dtype=np.complex128)
        for k in range(self.Nk):
            for n in range(self.Nb):
                pseudo_nkG[n, k] = \
                    calc.get_pseudo_wave_function(
                        band=n, kpt=k, spin=spin).ravel()

        # Use initial guess to determine U and C
        C_kul, U_kww = scdm(pseudo_nkG,
                            kpts=kpt_kc,
                            fixed_k=self.fixedstates_k,
                            Nw=self.Nw)
        return WannierState(C_kul, U_kww)


class WannierState:
    def __init__(self, C_kul, U_kww):
        # Number of u is not always the same, so C_kul is ragged
        self.C_kul = [C_ul.astype(complex) for C_ul in C_kul]
        self.U_kww = U_kww



import pytest

from ase import Atoms
from ase.calculators.test import FreeElectrons
from ase.lattice import all_variants
from ase.spectrum.band_structure import calculate_band_structure


@pytest.mark.parametrize("i, lat",
                         [pytest.param(i, lat, id=lat.variant)
                          for i, lat in enumerate(all_variants())
                          if lat.ndim == 3])
def test_lattice_bandstructure(testdir, i, lat, figure):
    xid = f'{i:02d}.{lat.variant}'
    path = lat.bandpath(density=10)
    path.write(f'path.{xid}.json')
    atoms = Atoms(cell=lat.tocell(), pbc=True)
    atoms.calc = FreeElectrons(nvalence=0, kpts=path.kpts)
    bs = calculate_band_structure(atoms, path)
    bs.write(f'bs.{xid}.json')

    ax = figure.gca()
    bs.plot(ax=ax, emin=0, emax=20, filename=f'fig.{xid}.png')



"""This is a module to handle generic ASE (gui) defaults ...

... from a ~/.ase/gui.py configuration file, if it exists. It is imported when
opening ASE-GUI and can then be modified at runtime, if necessary. syntax for
each entry:

gui_default_settings['key'] = value
"""
import runpy

gui_default_settings = {
    'gui_graphs_string': 'i, e - E[-1]',  # default for the graph command
    'gui_foreground_color': '#000000',
    'gui_background_color': '#ffffff',
    'covalent_radii': None,
    'radii_scale': 0.89,
    'force_vector_scale': 1.0,
    'velocity_vector_scale': 1.0,
    'show_unit_cell': True,
    'show_axes': True,
    'show_bonds': False,
    'shift_cell': False,
    'swap_mouse': False,
}


def read_defaults():
    import os

    # should look for .config/ase/gui.py
    # if 'XDG_CONFIG_HOME' in os.environ:
    #    name = os.environ['XDG_CONFIG_HOME'] + '/ase/gui.py'
    name = os.path.expanduser('~/.ase/gui.py')
    config = gui_default_settings
    if os.path.exists(name):
        runpy.run_path(name, init_globals={'gui_default_settings':
                                           gui_default_settings})
    return config



import os

import pytest

from ase.db import connect


@pytest.fixture(scope='session')
def mysql_port():
    return int(os.environ.get('MYSQL_TCP_PORT', 3306))


@pytest.fixture()
def get_db_name(mysql_port):
    """ Fixture that returns a function to get the test db name
    for the different supported db types.

    Args:
        dbtype (str): Type of database. Currently only 5 types supported:
            postgresql, mysql, mariadb, json, and db (sqlite3)
        clean_db (bool): Whether to clean all entries from the db. Useful
            for reusing the database across multiple tests. Defaults to True.
    """
    def _func(dbtype, clean_db=True):
        name = None

        if dbtype == 'postgresql':
            pytest.importorskip('psycopg2')
            if os.environ.get('POSTGRES_DB'):  # gitlab-ci
                name = 'postgresql://ase:ase@postgres:5432/testase'
            else:
                name = os.environ.get('ASE_TEST_POSTGRES_URL')
        elif dbtype == 'mysql':
            pytest.importorskip('pymysql')
            if os.environ.get('CI_PROJECT_DIR'):  # gitlab-ci
                # Note: testing of non-standard port by changing from default
                # of 3306 to 3307
                name = f'mysql://root:ase@mysql:{mysql_port}/testase_mysql'
            else:
                name = os.environ.get('MYSQL_DB_URL')
        elif dbtype == 'mariadb':
            pytest.importorskip('pymysql')
            if os.environ.get('CI_PROJECT_DIR'):  # gitlab-ci
                # Note: testing of non-standard port by changing from default
                # of 3306 to 3307
                name = f'mariadb://root:ase@mariadb:{mysql_port}/testase_mysql'
            else:
                name = os.environ.get('MYSQL_DB_URL')
        elif dbtype == 'json':
            name = 'testase.json'
        elif dbtype == 'db':
            name = 'testase.db'
        else:
            raise ValueError(f'Bad db type: {dbtype}')

        if name is None:
            pytest.skip('Test requires environment variables')

        if clean_db:
            if dbtype in ["postgresql", "mysql", "mariadb"]:
                c = connect(name)
                c.delete([row.id for row in c.select()])

        return name

    return _func


# For different parametrizations (we will move the heavier ones out anyway)
dbtypes = ['db', 'postgresql', 'mysql', 'mariadb']


@pytest.fixture(params=dbtypes)
def dbtype(request):
    return xfail_bad_backends(request)


@pytest.fixture(params=['json', *dbtypes])
def dbtype2(request):
    return xfail_bad_backends(request)


def xfail_bad_backends(request):
    if request.param in {'postgresql', 'mysql', 'mariadb'}:
        pytest.xfail(reason='race condition')
    return request.param



from pathlib import Path

import pytest

from ase.calculators.calculator import Calculator


def test_directory_and_label():
    def normalize(path):
        """Helper function to normalize path"""
        return str(Path(path))

    calc = Calculator()

    assert calc.directory == '.'
    assert calc.label is None

    calc.directory = 'somedir'

    assert calc.directory == 'somedir'
    assert calc.label == 'somedir/'

    # We cannot redundantly specify directory
    with pytest.raises(ValueError):
        calc = Calculator(directory='somedir',
                          label='anotherdir/label')

    # Test only directory in directory
    calc = Calculator(directory='somedir',
                      label='label')

    assert calc.directory == 'somedir'
    assert calc.label == 'somedir/label'

    wdir = '/home/somedir'
    calc = Calculator(directory=wdir,
                      label='label')

    assert calc.directory == normalize(wdir)
    assert calc.label == normalize(wdir) + '/label'

    # Test we can handle pathlib directories
    wdir = Path('/home/somedir')
    calc = Calculator(directory=wdir,
                      label='label')
    assert calc.directory == normalize(wdir)
    assert calc.label == normalize(wdir) + '/label'

    with pytest.raises(ValueError):
        calc = Calculator(directory=wdir,
                          label='somedir/label')

    # Passing in empty directories with directories in label should be OK
    for wdir in ['somedir', '/home/directory']:
        label = wdir + '/label'
        expected_label = normalize(wdir) + '/label'
        calc = Calculator(directory='', label=label)
        assert calc.label == expected_label
        assert calc.directory == normalize(wdir)

        calc = Calculator(directory='.', label=label)
        assert calc.label == expected_label
        assert calc.directory == normalize(wdir)


def test_deprecated_get_spin_polarized():
    calc = Calculator()
    with pytest.warns(FutureWarning):
        spinpol = calc.get_spin_polarized()
    assert spinpol is False



"""
ASE Calculator for interatomic models compatible with the Knowledgebase
of Interatomic Models (KIM) application programming interface (API).
Written by:

Mingjian Wen
Daniel S. Karls
University of Minnesota
"""
import numpy as np

from ase.calculators.calculator import Calculator, compare_atoms

from . import kimpy_wrappers, neighborlist


class KIMModelData:
    """Initializes and subsequently stores the KIM API Portable Model
    object, KIM API ComputeArguments object, and the neighbor list
    object used by instances of KIMModelCalculator.  Also stores the
    arrays which are registered in the KIM API and which are used to
    communicate with the model.
    """

    def __init__(self, model_name, ase_neigh, neigh_skin_ratio, debug=False):
        self.model_name = model_name
        self.ase_neigh = ase_neigh
        self.debug = debug

        # Initialize KIM API Portable Model object and ComputeArguments object
        self.kim_model, self.compute_args = self._init_kim()

        self.species_map = self._create_species_map()

        # Ask model to provide information relevant for neighbor list
        # construction
        (
            model_influence_dist,
            model_cutoffs,
            padding_not_require_neigh,
        ) = self.get_model_neighbor_list_parameters()

        # Initialize neighbor list object
        self.neigh = self._init_neigh(
            neigh_skin_ratio,
            model_influence_dist,
            model_cutoffs,
            padding_not_require_neigh,
        )

    def _init_kim(self):
        """Create the KIM API Portable Model object and KIM API ComputeArguments
        object
        """
        if self.kim_initialized:
            return None

        kim_model = kimpy_wrappers.PortableModel(self.model_name, self.debug)

        # KIM API model object is what actually creates/destroys the
        # ComputeArguments object, so we must pass it as a parameter
        compute_args = kim_model.compute_arguments_create()

        return kim_model, compute_args

    def _init_neigh(
        self,
        neigh_skin_ratio,
        model_influence_dist,
        model_cutoffs,
        padding_not_require_neigh,
    ):
        """Initialize neighbor list, either an ASE-native neighborlist
        or one created using the neighlist module in kimpy
        """
        neigh_list_object_type = (
            neighborlist.ASENeighborList
            if self.ase_neigh
            else neighborlist.KimpyNeighborList
        )
        return neigh_list_object_type(
            self.compute_args,
            neigh_skin_ratio,
            model_influence_dist,
            model_cutoffs,
            padding_not_require_neigh,
            self.debug,
        )

    def get_model_neighbor_list_parameters(self):
        model_influence_dist = self.kim_model.get_influence_distance()
        (
            model_cutoffs,
            padding_not_require_neigh,
        ) = self.kim_model.get_neighbor_list_cutoffs_and_hints()

        return model_influence_dist, model_cutoffs, padding_not_require_neigh

    def update_compute_args_pointers(self, energy, forces):
        self.compute_args.update(
            self.num_particles,
            self.species_code,
            self._particle_contributing,
            self.coords,
            energy,
            forces,
        )

    def _create_species_map(self):
        """Get all the supported species of the KIM model and the
        corresponding integer codes used by the model

        Returns
        -------
        species_map : dict
            key : str
                chemical symbols (e.g. "Ar")
            value : int
                species integer code (e.g. 1)
        """
        supported_species, codes = self._get_model_supported_species_and_codes()
        species_map = {}
        for i, spec in enumerate(supported_species):
            species_map[spec] = codes[i]
            if self.debug:
                print(
                    "Species {} is supported and its code is: {}".format(
                        spec, codes[i])
                )

        return species_map

    @property
    def padding_image_of(self):
        return self.neigh.padding_image_of

    @property
    def num_particles(self):
        return self.neigh.num_particles

    @property
    def coords(self):
        return self.neigh.coords

    @property
    def _particle_contributing(self):
        return self.neigh.particle_contributing

    @property
    def species_code(self):
        return self.neigh.species_code

    @property
    def kim_initialized(self):
        return hasattr(self, "kim_model")

    @property
    def _neigh_initialized(self):
        return hasattr(self, "neigh")

    @property
    def _get_model_supported_species_and_codes(self):
        return self.kim_model.get_model_supported_species_and_codes


class KIMModelCalculator(Calculator):
    """Calculator that works with KIM Portable Models (PMs).

    Calculator that carries out direct communication between ASE and a
    KIM Portable Model (PM) through the kimpy library (which provides a
    set of python bindings to the KIM API).

    Parameters
    ----------
    model_name : str
      The unique identifier assigned to the interatomic model (for
      details, see https://openkim.org/doc/schema/kim-ids)

    ase_neigh : bool, optional
      False (default): Use kimpy's neighbor list library

      True: Use ASE's internal neighbor list mechanism (usually slower
      than the kimpy neighlist library)

    neigh_skin_ratio : float, optional
      Used to determine the neighbor list cutoff distance, r_neigh,
      through the relation r_neigh = (1 + neigh_skin_ratio) * rcut,
      where rcut is the model's influence distance. (Default: 0.2)

    release_GIL : bool, optional
      Whether to release python GIL.  Releasing the GIL allows a KIM
      model to run with multiple concurrent threads. (Default: False)

    debug : bool, optional
      If True, detailed information is printed to stdout. (Default:
      False)
    """

    implemented_properties = ["energy", "free_energy", "forces", "stress"]

    ignored_changes = {"initial_charges", "initial_magmoms"}

    def __init__(
        self,
        model_name,
        ase_neigh=False,
        neigh_skin_ratio=0.2,
        release_GIL=False,
        debug=False,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)

        self.model_name = model_name
        self.release_GIL = release_GIL
        self.debug = debug

        if neigh_skin_ratio < 0:
            raise ValueError('Argument "neigh_skin_ratio" must be non-negative')
        self.neigh_skin_ratio = neigh_skin_ratio

        # Model output
        self.energy = None
        self.forces = None

        # Create KIMModelData object. This will take care of creating
        # and storing the KIM API Portable Model object, KIM API
        # ComputeArguments object, and the neighbor list object that
        # our calculator needs
        self._kimmodeldata = KIMModelData(
            self.model_name, ase_neigh, self.neigh_skin_ratio, self.debug
        )

        self._parameters_changed = False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, value, traceback):
        pass

    def __repr__(self):
        return f"KIMModelCalculator(model_name={self.model_name})"

    def calculate(
        self,
        atoms=None,
        properties=["energy", "forces", "stress"],
        system_changes=["positions", "numbers", "cell", "pbc"],
    ):
        """
        Inherited method from the ase Calculator class that is called by
        get_property()

        Parameters
        ----------
        atoms : Atoms
            Atoms object whose properties are desired

        properties : list of str
            List of what needs to be calculated.  Can be any combination
            of 'energy', 'forces' and 'stress'.

        system_changes : list of str
            List of what has changed since last calculation.  Can be any
            combination of these six: 'positions', 'numbers', 'cell',
            and 'pbc'.
        """

        super().calculate(atoms, properties, system_changes)

        if self._parameters_changed:
            self._parameters_changed = False

        if system_changes:

            # Ask model to update all of its parameters and the parameters
            # related to the neighbor list(s). This update is necessary to do
            # here since the user will generally have made changes the model
            # parameters since the last time an update was performed and we
            # need to ensure that any properties calculated here are made using
            # the up-to-date model and neighbor list parameters.
            self._model_refresh_and_update_neighbor_list_parameters()

            if self._need_neigh_update(atoms, system_changes):
                self._update_neigh(atoms, self._species_map)
                self.energy = np.array([0.0], dtype=kimpy_wrappers.c_double)
                self.forces = np.zeros(
                    [self._num_particles[0], 3], dtype=kimpy_wrappers.c_double
                )
                self._update_compute_args_pointers(self.energy, self.forces)
            else:
                self._update_kim_coords(atoms)

            self._kim_model.compute(self._compute_args, self.release_GIL)

        energy = self.energy[0]
        forces = self._assemble_padding_forces()

        try:
            volume = atoms.get_volume()
            stress = self._compute_virial_stress(
                self.forces, self._coords, volume)
        except ValueError:  # Volume cannot be computed
            stress = None

        # Quantities passed back to ASE
        self.results["energy"] = energy
        self.results["free_energy"] = energy
        self.results["forces"] = forces
        self.results["stress"] = stress

    def check_state(self, atoms, tol=1e-15):
        # Check for change in atomic configuration (positions or pbc)
        system_changes = compare_atoms(
            self.atoms, atoms, excluded_properties=self.ignored_changes
        )

        # Check if model parameters were changed
        if self._parameters_changed:
            system_changes.append("calculator")

        return system_changes

    def _assemble_padding_forces(self):
        """
        Assemble forces on padding atoms back to contributing atoms.

        Parameters
        ----------
        forces : 2D array of doubles
            Forces on both contributing and padding atoms

        num_contrib:  int
            Number of contributing atoms

        padding_image_of : 1D array of int
            Atom number, of which the padding atom is an image


        Returns
        -------
            Total forces on contributing atoms.
        """

        total_forces = np.array(self.forces[:self._num_contributing_particles])

        if self._padding_image_of.size != 0:
            pad_forces = self.forces[self._num_contributing_particles:]
            for f, org_index in zip(pad_forces, self._padding_image_of):
                total_forces[org_index] += f

        return total_forces

    @staticmethod
    def _compute_virial_stress(forces, coords, volume):
        """Compute the virial stress in Voigt notation.

        Parameters
        ----------
        forces : 2D array
            Partial forces on all atoms (padding included)

        coords : 2D array
            Coordinates of all atoms (padding included)

        volume : float
            Volume of cell

        Returns
        -------
        stress : 1D array
            stress in Voigt order (xx, yy, zz, yz, xz, xy)
        """
        stress = np.zeros(6)
        stress[0] = -np.dot(forces[:, 0], coords[:, 0]) / volume
        stress[1] = -np.dot(forces[:, 1], coords[:, 1]) / volume
        stress[2] = -np.dot(forces[:, 2], coords[:, 2]) / volume
        stress[3] = -np.dot(forces[:, 1], coords[:, 2]) / volume
        stress[4] = -np.dot(forces[:, 0], coords[:, 2]) / volume
        stress[5] = -np.dot(forces[:, 0], coords[:, 1]) / volume

        return stress

    @property
    def _update_compute_args_pointers(self):
        return self._kimmodeldata.update_compute_args_pointers

    @property
    def _kim_model(self):
        return self._kimmodeldata.kim_model

    @property
    def _compute_args(self):
        return self._kimmodeldata.compute_args

    @property
    def _num_particles(self):
        return self._kimmodeldata.num_particles

    @property
    def _coords(self):
        return self._kimmodeldata.coords

    @property
    def _padding_image_of(self):
        return self._kimmodeldata.padding_image_of

    @property
    def _species_map(self):
        return self._kimmodeldata.species_map

    @property
    def _neigh(self):
        # WARNING: This property is underscored for a reason! The
        # neighborlist(s) itself (themselves) may not be up to date with
        # respect to changes that have been made to the model's parameters, or
        # even since the positions in the Atoms object may have changed.
        # Neighbor lists are only potentially updated inside the ``calculate``
        # method.
        return self._kimmodeldata.neigh

    @property
    def _num_contributing_particles(self):
        return self._neigh.num_contributing_particles

    @property
    def _update_kim_coords(self):
        return self._neigh.update_kim_coords

    @property
    def _need_neigh_update(self):
        return self._neigh.need_neigh_update

    @property
    def _update_neigh(self):
        return self._neigh.update

    @property
    def parameters_metadata(self):
        return self._kim_model.parameters_metadata

    @property
    def parameter_names(self):
        return self._kim_model.parameter_names

    @property
    def get_parameters(self):
        # Ask model to update all of its parameters and the parameters related
        # to the neighbor list(s). This update is necessary to do here since
        # the user will generally have made changes the model parameters since
        # the last time an update was performed and we need to ensure the
        # parameters returned by this method are fully up to date.
        self._model_refresh_and_update_neighbor_list_parameters()

        return self._kim_model.get_parameters

    def set_parameters(self, **kwargs):
        parameters = self._kim_model.set_parameters(**kwargs)
        self._parameters_changed = True

        return parameters

    def _model_refresh_and_update_neighbor_list_parameters(self):
        """
        Call the model's refresh routine and update the neighbor list object
        for any necessary changes arising from changes to the model parameters,
        e.g. a change in one of its cutoffs.  After a model's parameters have
        been changed, this method *must* be called before calling the model's
        compute routine.
        """
        self._kim_model.clear_then_refresh()

        # Update neighbor list parameters
        (
            model_influence_dist,
            model_cutoffs,
            padding_not_require_neigh,
        ) = self._kimmodeldata.get_model_neighbor_list_parameters()

        self._neigh.set_neigh_parameters(
            self.neigh_skin_ratio,
            model_influence_dist,
            model_cutoffs,
            padding_not_require_neigh,
        )



import numpy as np

from ase.cluster.cubic import SimpleCubicFactory


# The L1_2 structure is "based on FCC", but is really simple cubic
# with a basis.
class AuCu3Factory(SimpleCubicFactory):
    "A factory for creating AuCu3 (L1_2) lattices."

    atomic_basis = np.array([[0., 0., 0.],
                             [0., .5, .5],
                             [.5, 0., .5],
                             [.5, .5, 0.]])

    element_basis = [0, 1, 1, 1]


AuCu3 = L1_2 = AuCu3Factory()



import numpy as np
import numpy.linalg as la


class Kernel():
    def __init__(self):
        pass

    def set_params(self, params):
        pass

    def kernel(self, x1, x2):
        """Kernel function to be fed to the Kernel matrix"""

    def K(self, X1, X2):
        """Compute the kernel matrix """
        return np.block([[self.kernel(x1, x2) for x2 in X2] for x1 in X1])


class SE_kernel(Kernel):
    """Squared exponential kernel without derivatives"""

    def __init__(self):
        Kernel.__init__(self)

    def set_params(self, params):
        """Set the parameters of the squared exponential kernel.

        Parameters:

        params: [weight, l] Parameters of the kernel:
            weight: prefactor of the exponential
            l : scale of the kernel
        """
        self.weight = params[0]
        self.l = params[1]

    def squared_distance(self, x1, x2):
        """Returns the norm of x1-x2 using diag(l) as metric """
        return np.sum((x1 - x2) * (x1 - x2)) / self.l**2

    def kernel(self, x1, x2):
        """ This is the squared exponential function"""
        return self.weight**2 * np.exp(-0.5 * self.squared_distance(x1, x2))

    def dK_dweight(self, x1, x2):
        """Derivative of the kernel respect to the weight """
        return 2 * self.weight * np.exp(-0.5 * self.squared_distance(x1, x2))

    def dK_dl(self, x1, x2):
        """Derivative of the kernel respect to the scale"""
        return self.kernel * la.norm(x1 - x2)**2 / self.l**3


class SquaredExponential(SE_kernel):
    """Squared exponential kernel with derivatives.
    For the formulas see Koistinen, Dagbjartsdottir, Asgeirsson, Vehtari,
    Jonsson.
    Nudged elastic band calculations accelerated with Gaussian process
    regression. Section 3.

    Before making any predictions, the parameters need to be set using the
    method SquaredExponential.set_params(params) where the parameters are a
    list whose first entry is the weight (prefactor of the exponential) and
    the second is the scale (l).

    Parameters:

    dimensionality: The dimensionality of the problem to optimize, typically
                    3*N where N is the number of atoms. If dimensionality is
                    None, it is computed when the kernel method is called.

    Attributes:
    ----------------
    D:          int. Dimensionality of the problem to optimize
    weight:     float. Multiplicative constant to the exponenetial kernel
    l :         float. Length scale of the squared exponential kernel

    Relevant Methods:
    ----------------
    set_params:         Set the parameters of the Kernel, i.e. change the
                        attributes
    kernel_function:    Squared exponential covariance function
    kernel:             covariance matrix between two points in the manifold.
                        Note that the inputs are arrays of shape (D,)
    kernel_matrix:      Kernel matrix of a data set to itself, K(X,X)
                        Note the input is an array of shape (nsamples, D)
    kernel_vector       Kernel matrix of a point x to a dataset X, K(x,X).

    gradient:           Gradient of K(X,X) with respect to the parameters of
                        the kernel i.e. the hyperparameters of the Gaussian
                        process.
    """

    def __init__(self, dimensionality=None):
        self.D = dimensionality
        SE_kernel.__init__(self)

    def kernel_function(self, x1, x2):
        """ This is the squared exponential function"""
        return self.weight**2 * np.exp(-0.5 * self.squared_distance(x1, x2))

    def kernel_function_gradient(self, x1, x2):
        """Gradient of kernel_function respect to the second entry.
        x1: first data point
        x2: second data point
        """
        prefactor = (x1 - x2) / self.l**2
        # return prefactor * self.kernel_function(x1,x2)
        return prefactor

    def kernel_function_hessian(self, x1, x2):
        """Second derivatives matrix of the kernel function"""
        P = np.outer(x1 - x2, x1 - x2) / self.l**2
        prefactor = (np.identity(self.D) - P) / self.l**2
        return prefactor

    def kernel(self, x1, x2):
        """Squared exponential kernel including derivatives.
        This function returns a D+1 x D+1 matrix, where D is the dimension of
        the manifold.
        """
        K = np.identity(self.D + 1)
        K[0, 1:] = self.kernel_function_gradient(x1, x2)
        K[1:, 0] = -K[0, 1:]
        # K[1:,1:] = self.kernel_function_hessian(x1, x2)
        P = np.outer(x1 - x2, x1 - x2) / self.l**2
        K[1:, 1:] = (K[1:, 1:] - P) / self.l**2
        # return np.block([[k,j2],[j1,h]])*self.kernel_function(x1, x2)
        return K * self.kernel_function(x1, x2)

    def kernel_matrix(self, X):
        """This is the same method than self.K for X1=X2, but using the matrix
        is then symmetric.
        """
        n, D = np.atleast_2d(X).shape
        K = np.identity(n * (D + 1))
        self.D = D
        D1 = D + 1

        # fill upper triangular:
        for i in range(n):
            for j in range(i + 1, n):
                k = self.kernel(X[i], X[j])
                K[i * D1:(i + 1) * D1, j * D1:(j + 1) * D1] = k
                K[j * D1:(j + 1) * D1, i * D1:(i + 1) * D1] = k.T
            K[i * D1:(i + 1) * D1, i * D1:(i + 1) * D1] = self.kernel(
                X[i], X[i])
        return K

    def kernel_vector(self, x, X, nsample):
        return np.hstack([self.kernel(x, x2) for x2 in X])

    # ---------Derivatives--------
    def dK_dweight(self, X):
        """Return the derivative of K(X,X) respect to the weight """
        return self.K(X, X) * 2 / self.weight

    # ----Derivatives of the kernel function respect to the scale ---
    def dK_dl_k(self, x1, x2):
        """Returns the derivative of the kernel function respect to l"""
        return self.squared_distance(x1, x2) / self.l

    def dK_dl_j(self, x1, x2):
        """Returns the derivative of the gradient of the kernel function
        respect to l
        """
        prefactor = -2 * (1 - 0.5 * self.squared_distance(x1, x2)) / self.l
        return self.kernel_function_gradient(x1, x2) * prefactor

    def dK_dl_h(self, x1, x2):
        """Returns the derivative of the hessian of the kernel function respect
        to l
        """
        I = np.identity(self.D)
        P = np.outer(x1 - x2, x1 - x2) / self.l**2
        prefactor = 1 - 0.5 * self.squared_distance(x1, x2)
        return -2 * (prefactor * (I - P) - P) / self.l**3

    def dK_dl_matrix(self, x1, x2):
        k = np.asarray(self.dK_dl_k(x1, x2)).reshape((1, 1))
        j2 = self.dK_dl_j(x1, x2).reshape(1, -1)
        j1 = self.dK_dl_j(x2, x1).reshape(-1, 1)
        h = self.dK_dl_h(x1, x2)
        return np.block([[k, j2], [j1, h]]) * self.kernel_function(x1, x2)

    def dK_dl(self, X):
        """Return the derivative of K(X,X) respect of l"""
        return np.block([[self.dK_dl_matrix(x1, x2) for x2 in X] for x1 in X])

    def gradient(self, X):
        """Computes the gradient of matrix K given the data respect to the
        hyperparameters. Note matrix K here is self.K(X,X).
        Returns a 2-entry list of n(D+1) x n(D+1) matrices
        """
        return [self.dK_dweight(X), self.dK_dl(X)]



import numpy as np
import pytest

from ase import Atoms
from ase.calculators.bond_polarizability import (
    BondPolarizability,
    Linearized,
    LippincottStuttman,
)


def test_CC_bond():
    """Test polarizabilties of a single CC bond"""
    C2 = Atoms('C2', positions=[[0, 0, 0], [0, 0, 1.69]])

    def check_symmetry(alpha):
        alpha_diag = np.diagonal(alpha)
        assert alpha == pytest.approx(np.diag(alpha_diag))
        assert alpha_diag[0] == alpha_diag[1]

    bp = BondPolarizability()
    check_symmetry(bp(C2))
    bp = BondPolarizability(Linearized())
    check_symmetry(bp(C2))


def test_symmetry():
    for lin in [LippincottStuttman(), Linearized()]:
        assert lin('B', 'N', 1) == lin('N', 'B', 1)


def test_2to3():
    """Compare polarizabilties of one and two bonds"""
    Si2 = Atoms('Si2', positions=[[0, 0, 0], [0, 0, 2.5]])
    Si3 = Atoms('Si3', positions=[[0, 0, -2.5], [0, 0, 0], [0, 0, 2.5]])
    bp = BondPolarizability()
    bp2 = bp(Si2)
    # polarizability is a tensor
    assert bp2.shape == (3, 3)
    # check sum of equal bonds
    assert bp(Si3) == pytest.approx(2 * bp2)



# ######################################
# Implementation of FIRE2.0 and ABC-FIRE

# The FIRE2.0 algorithm is implemented using the integrator euler semi implicit
#  as described in the paper:
#   J. Guénolé, W.G. Nöhring, A. Vaid, F. Houllé, Z. Xie, A. Prakash,
#   E. Bitzek,
#    Assessment and optimization of the fast inertial relaxation engine (fire)
#    for energy minimization in atomistic simulations and its
#    implementation in lammps,
#    Comput. Mater. Sci. 175 (2020) 109584.
#    https://doi.org/10.1016/j.commatsci.2020.109584.
#    This implementation does not include N(p<0), initialdelay
#
# ABC-Fire is implemented as described in the paper:
#   S. Echeverri Restrepo, P. Andric,
#    ABC-FIRE: Accelerated Bias-Corrected Fast Inertial Relaxation Engine,
#    Comput. Mater. Sci. 218 (2023) 111978.
#    https://doi.org/10.1016/j.commatsci.2022.111978.
#######################################

from typing import IO, Callable, Optional, Union

import numpy as np

from ase import Atoms
from ase.optimize.optimize import Optimizer


class FIRE2(Optimizer):
    def __init__(
        self,
        atoms: Atoms,
        restart: Optional[str] = None,
        logfile: Union[IO, str] = '-',
        trajectory: Optional[str] = None,
        dt: float = 0.1,
        maxstep: float = 0.2,
        dtmax: float = 1.0,
        dtmin: float = 2e-3,
        Nmin: int = 20,
        finc: float = 1.1,
        fdec: float = 0.5,
        astart: float = 0.25,
        fa: float = 0.99,
        position_reset_callback: Optional[Callable] = None,
        use_abc: Optional[bool] = False,
        **kwargs,
    ):
        """

        Parameters
        ----------
        atoms: :class:`~ase.Atoms`
            The Atoms object to relax.

        restart: str
            JSON file used to store hessian matrix. If set, file with
            such a name will be searched and hessian matrix stored will
            be used, if the file exists.

        logfile: file object or str
            If *logfile* is a string, a file with that name will be opened.
            Use '-' for stdout.

        trajectory: str
            Trajectory file used to store optimisation path.

        dt: float
            Initial time step. Defualt value is 0.1

        maxstep: float
            Used to set the maximum distance an atom can move per
            iteration (default value is 0.2). Note that for ABC-FIRE the
            check is done independently for each cartesian direction.

        dtmax: float
            Maximum time step. Default value is 1.0

        dtmin: float
            Minimum time step. Default value is 2e-3

        Nmin: int
            Number of steps to wait after the last time the dot product of
            the velocity and force is negative (P in The FIRE article) before
            increasing the time step. Default value is 20.

        finc: float
            Factor to increase the time step. Default value is 1.1

        fdec: float
            Factor to decrease the time step. Default value is 0.5

        astart: float
            Initial value of the parameter a. a is the Coefficient for
            mixing the velocity and the force. Called alpha in the FIRE article.
            Default value 0.25.

        fa: float
            Factor to decrease the parameter alpha. Default value is 0.99

        position_reset_callback: function(atoms, r, e, e_last)
            Function that takes current *atoms* object, an array of position
            *r* that the optimizer will revert to, current energy *e* and
            energy of last step *e_last*. This is only called if e > e_last.

        use_abc: bool
            If True, the Accelerated Bias-Corrected FIRE algorithm is
            used (ABC-FIRE).
            Default value is False.

        kwargs : dict, optional
            Extra arguments passed to
            :class:`~ase.optimize.optimize.Optimizer`.

       """
        Optimizer.__init__(self, atoms, restart, logfile, trajectory, **kwargs)

        self.dt = dt

        self.Nsteps = 0

        if maxstep is not None:
            self.maxstep = maxstep
        else:
            self.maxstep = self.defaults["maxstep"]

        self.dtmax = dtmax
        self.dtmin = dtmin
        self.Nmin = Nmin
        self.finc = finc
        self.fdec = fdec
        self.astart = astart
        self.fa = fa
        self.a = astart
        self.position_reset_callback = position_reset_callback
        self.use_abc = use_abc

    def initialize(self):
        self.v = None

    def read(self):
        self.v, self.dt = self.load()

    def step(self, f=None):
        optimizable = self.optimizable

        if f is None:
            f = optimizable.get_forces()

        if self.v is None:
            self.v = np.zeros((len(optimizable), 3))
        else:

            vf = np.vdot(f, self.v)
            if vf > 0.0:

                self.Nsteps += 1
                if self.Nsteps > self.Nmin:
                    self.dt = min(self.dt * self.finc, self.dtmax)
                    self.a *= self.fa
            else:
                self.Nsteps = 0
                self.dt = max(self.dt * self.fdec, self.dtmin)
                self.a = self.astart

                dr = - 0.5 * self.dt * self.v
                r = optimizable.get_positions()
                optimizable.set_positions(r + dr)
                self.v[:] *= 0.0

        # euler semi implicit
        f = optimizable.get_forces()
        self.v += self.dt * f

        if self.use_abc:
            self.a = max(self.a, 1e-10)
            abc_multiplier = 1. / (1. - (1. - self.a)**(self.Nsteps + 1))
            v_mix = ((1.0 - self.a) * self.v + self.a * f / np.sqrt(
                np.vdot(f, f)) * np.sqrt(np.vdot(self.v, self.v)))
            self.v = abc_multiplier * v_mix

            # Verifying if the maximum distance an atom
            #  moved is larger than maxstep, for ABC-FIRE the check
            #  is done independently for each cartesian direction
            if np.all(self.v):
                v_tmp = []
                for car_dir in range(3):
                    v_i = np.where(np.abs(self.v[:, car_dir]) *
                                   self.dt > self.maxstep,
                                   (self.maxstep / self.dt) *
                                   (self.v[:, car_dir] /
                                   np.abs(self.v[:, car_dir])),
                                   self.v[:, car_dir])
                    v_tmp.append(v_i)
                self.v = np.array(v_tmp).T

        else:
            self.v = ((1.0 - self.a) * self.v + self.a * f / np.sqrt(
                np.vdot(f, f)) * np.sqrt(np.vdot(self.v, self.v)))

        dr = self.dt * self.v

        # Verifying if the maximum distance an atom moved
        #  step is larger than maxstep, for FIRE2.
        if not self.use_abc:
            normdr = np.sqrt(np.vdot(dr, dr))
            if normdr > self.maxstep:
                dr = self.maxstep * dr / normdr

        r = optimizable.get_positions()
        optimizable.set_positions(r + dr)

        self.dump((self.v, self.dt))



import subprocess

import numpy as np

from ase import Atoms
from ase.calculators.amber import Amber
from ase.io import write


def test_amber(factories):
    """Test that amber calculator works.

    This is conditional on the existence of the $AMBERHOME/bin/sander
    executable.
    """

    factories.require('amber')

    with open('mm.in', 'w') as outfile:
        outfile.write("""\
    zero step md to get energy and force
    &cntrl
    imin=0, nstlim=0,  ntx=1 !0 step md
    cut=100, ntb=0,          !non-periodic
    ntpr=1,ntwf=1,ntwe=1,ntwx=1 ! (output frequencies)
    &end
    END
    """)

    atoms = Atoms('OH2OH2',
                  [[-0.956, -0.121, 0],
                   [-1.308, 0.770, 0],
                   [0.000, 0.000, 0],
                   [3.903, 0.000, 0],
                   [4.215, -0.497, -0.759],
                   [4.215, -0.497, 0.759]])
    atoms.arrays['residuenames'] = ['WAT'] * 6
    atoms.arrays['residuenumbers'] = [1] * 3 + [2] * 3
    atoms.arrays['atomtypes'] = ['O', 'H1', 'H2'] * 2
    write('2h2o.pdb', atoms)

    with open('tleap.in', 'w') as outfile:
        outfile.write("""\
    source leaprc.protein.ff14SB
    source leaprc.gaff
    source leaprc.water.tip3p
    mol = loadpdb 2h2o.pdb
    saveamberparm mol 2h2o.top 2h2o.inpcrd
    quit
    """)

    subprocess.call('tleap -f tleap.in'.split())

    calc = Amber(amber_exe='sander -O ',
                 infile='mm.in',
                 outfile='mm.out',
                 topologyfile='2h2o.top',
                 incoordfile='mm.crd')
    calc.write_coordinates(atoms, 'mm.crd')
    atoms_check = atoms.copy()
    calc.read_coordinates(atoms_check, 'mm.crd')
    assert np.allclose(atoms_check.positions, atoms.positions)
    assert np.allclose(atoms_check.get_velocities(), atoms.get_velocities())

    atoms.calc = calc

    e = atoms.get_potential_energy()
    assert abs(e + 0.046799672) < 5e-3



"""Tests for `surface`"""
import math

import numpy as np
import pytest

from ase import Atom, Atoms
from ase.build import (
    add_adsorbate,
    bulk,
    fcc111,
    fcc211,
    graphene,
    mx2,
    surface,
)


def test_surface():
    """Test general"""
    atoms = fcc211('Au', (3, 5, 8), vacuum=10.)
    assert len(atoms) == 120

    atoms = atoms.repeat((2, 1, 1))
    assert np.allclose(atoms.get_distance(0, 130), 2.88499566724)

    atoms = fcc111('Ni', (2, 2, 4), orthogonal=True)
    add_adsorbate(atoms, 'H', 1, 'bridge')
    add_adsorbate(atoms, Atom('O'), 1, 'fcc')
    add_adsorbate(atoms, Atoms('F'), 1, 'hcp')

    # The next test ensures that a simple string of multiple atoms
    # cannot be used, which should fail with a KeyError that reports
    # the name of the molecule due to the string failing to work with
    # Atom().
    failed = False
    try:
        add_adsorbate(atoms, 'CN', 1, 'ontop')
    except KeyError as err:
        failed = True
        assert err.args[0] == 'CN'
    assert failed

    # This test ensures that the default periodic behavior remains unchanged
    cubic_fcc = bulk("Al", a=4.05, cubic=True)
    surface_fcc = surface(cubic_fcc, (1, 1, 1), 3)

    assert list(surface_fcc.pbc) == [True, True, False]
    assert surface_fcc.cell[2][2] == 0

    # This test checks the new periodic option
    cubic_fcc = bulk("Al", a=4.05, cubic=True)
    surface_fcc = surface(cubic_fcc, (1, 1, 1), 3, periodic=True)

    assert (list(surface_fcc.pbc) == [True, True, True])
    expected_length = 4.05 * 3**0.5  # for FCC with a=4.05
    assert math.isclose(surface_fcc.cell[2][2], expected_length)

    # This test checks the tags
    print(surface_fcc.get_tags())
    tags = np.array([3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 1])
    np.testing.assert_array_equal(surface_fcc.get_tags(), tags)


@pytest.mark.parametrize("vacuum", [None, 10.0])
def test_others(vacuum):
    """Test if other types of `surface` functions (at least) run."""
    mx2(kind='2H', vacuum=vacuum)
    mx2(kind='1T', vacuum=vacuum)
    graphene(vacuum=vacuum)
    graphene(thickness=0.5, vacuum=vacuum)



import numpy as np


class FixRotation:
    """Remove rotation from an atoms object.

    This class is intended as an observer on an atoms class during
    a molecular dynamics simulation.  When it is called, it removes
    any rotation around the center of mass.

    It assumes that the system is a (nano)particle with free boundary
    conditions.

    Bugs:
    Should check that the boundary conditions make sense.
    """

    def __init__(self, atoms):
        self.atoms = atoms

    def __call__(self):
        atoms = self.atoms

        r = atoms.get_positions() - atoms.get_center_of_mass()
        v = atoms.get_velocities()
        p = atoms.get_momenta()
        m = atoms.get_masses()

        x = r[:, 0]
        y = r[:, 1]
        z = r[:, 2]

        I11 = np.sum(m * (y**2 + z**2))
        I22 = np.sum(m * (x**2 + z**2))
        I33 = np.sum(m * (x**2 + y**2))
        I12 = np.sum(-m * x * y)
        I13 = np.sum(-m * x * z)
        I23 = np.sum(-m * y * z)

        I = np.array([[I11, I12, I13],
                      [I12, I22, I23],
                      [I13, I23, I33]])

        w = np.dot(np.linalg.inv(I), np.sum(np.cross(r, p), axis=0))

        self.atoms.set_velocities(v - np.cross(w, r))



from typing import Optional

import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure


class SimplePlottingAxes:
    def __init__(self,
                 ax: Optional[Axes] = None,
                 show: bool = False,
                 filename: str = None) -> None:
        self.ax = ax
        self.show = show
        self.filename = filename
        self.figure: Optional[Figure] = None

    def __enter__(self) -> Axes:
        if self.ax is None:
            self.figure, self.ax = plt.subplots()
        else:
            self.figure = self.ax.get_figure()

        return self.ax

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if exc_type is None:
            # If there was no exception, display/write the plot as appropriate
            if self.figure is None:
                raise Exception("Something went wrong initializing matplotlib "
                                "figure")
            if self.show:
                self.figure.show()
            if self.filename is not None:
                self.figure.savefig(self.filename)

        return



"""
Function-like objects that creates cubic clusters.
"""

import numpy as np

from ase.cluster.factory import ClusterFactory
from ase.data import reference_states as _refstate


class HexagonalFactory(ClusterFactory):
    spacegroup = 191

    xtal_name = 'hexagonal'

    def get_lattice_constant(self):
        "Get the lattice constant of an element with cubic crystal structure."
        symmetry = _refstate[self.atomic_numbers[0]]['symmetry']
        if symmetry != self.xtal_name:
            raise ValueError(f"Cannot guess the {self.xtal_name} " +
                             "lattice constant of an element with crystal " +
                             f"structure {symmetry}.")
        return _refstate[self.atomic_numbers[0]].copy()

    def set_basis(self):
        lattice = self.lattice_constant
        if isinstance(lattice, dict):
            a = lattice['a']
            try:
                c = lattice['c']
            except KeyError:
                c = a * lattice['c/a']
        else:
            if len(lattice) == 2:
                (a, c) = lattice
            else:
                raise ValueError(
                    f"Improper lattice constants for {self.xtal_name} crystal.")

        self.lattice_constant = (a, c)
        self.lattice_basis = np.array([[a, 0., 0.],
                                       [-a / 2., a * np.sqrt(3.) / 2., 0.],
                                       [0., 0., c]])
        self.resiproc_basis = self.get_resiproc_basis(self.lattice_basis)

    def set_surfaces_layers(self, surfaces, layers):
        for i, s in enumerate(surfaces):
            if len(s) == 4:
                (a, b, c, d) = s
                if a + b + c != 0:
                    raise ValueError(
                        "(%d,%d,%d,%d) is not a valid hexagonal Miller "
                        "index, as the sum of the first three numbers "
                        "should be zero." % (a, b, c, d))
                surfaces[i] = [a, b, d]

        ClusterFactory.set_surfaces_layers(self, surfaces, layers)


Hexagonal = HexagonalFactory()


class HexagonalClosedPackedFactory(HexagonalFactory):
    """A factory for creating HCP clusters."""
    spacegroup = 194

    xtal_name = 'hcp'

    atomic_basis = np.array([[0., 0., 0.],
                             [1. / 3., 2. / 3., .5]])


HexagonalClosedPacked = HexagonalClosedPackedFactory()


class GraphiteFactory(HexagonalFactory):
    """A factory for creating graphite clusters."""
    xtal_name = "graphite"

    atomic_basis = np.array([[0., 0., 0.],
                             [1. / 3., 2. / 3., 0.],
                             [1. / 3., 2. / 3., .5],
                             [2. / 3., 1. / 3., .5]])


Graphite = GraphiteFactory()



"""Temporary file while we deprecate this location."""

from ase.mep import NEB as RealNEB
from ase.mep import NEBTools as RealNEBTools
from ase.mep import idpp_interpolate as realidpp_interpolate
from ase.mep import interpolate as realinterpolate
from ase.utils import deprecated


class NEB(RealNEB):
    @deprecated('Please import NEB from ase.mep, not ase.neb.')
    def __init__(self, *args, **kwargs):
        """
        .. deprecated:: 3.23.0
            Please import :class:`~ase.mep.neb.NEB` from :mod:`ase.mep`
        """
        super().__init__(*args, **kwargs)


class NEBTools(RealNEBTools):
    @deprecated('Please import NEBTools from ase.mep, not ase.neb.')
    def __init__(self, *args, **kwargs):
        """
        .. deprecated:: 3.23.0
            Please import :class:`~ase.mep.neb.NEBTools`` from :mod:`ase.mep`
        """
        super().__init__(*args, **kwargs)


@deprecated('Please import interpolate from ase.mep, not ase.neb.')
def interpolate(*args, **kwargs):
    """
    .. deprecated:: 3.23.0
            Please import :func:`~ase.mep.neb.interpolate` from :mod:`ase.mep`
    """
    return realinterpolate(*args, **kwargs)


@deprecated('Please import idpp_interpolate from ase.mep, not ase.neb.')
def idpp_interpolate(*args, **kwargs):
    """
    .. deprecated:: 3.23.0
            Please import :func:`~ase.mep.neb.idpp_interpolate` from
            :mod:`ase.mep`
    """
    return realidpp_interpolate(*args, **kwargs)



import io
import warnings

import numpy as np
import pytest

from ase import Atoms
from ase.build import molecule
from ase.calculators.calculator import compare_atoms
from ase.io.bytes import parse_atoms, to_bytes
from ase.io.cif import (
    CIFLoop,
    NoStructureData,
    parse_cif,
    parse_loop,
    read_cif,
    write_cif,
)


def parse_string(string):
    buf = io.BytesIO(string.encode('latin1'))
    blocks = list(parse_cif(buf))
    return blocks


def check_fractional_occupancies(atoms):
    """ Checks fractional occupancy entries in atoms.info dict """
    assert atoms.info['occupancy']
    assert list(atoms.arrays['spacegroup_kinds'])

    occupancies = atoms.info['occupancy']
    for key in occupancies:
        assert isinstance(key, str)

    kinds = atoms.arrays['spacegroup_kinds']
    for a in atoms:
        a_index_str = str(kinds[a.index])
        if a.symbol == 'Na':

            assert len(occupancies[a_index_str]) == 2
            assert occupancies[a_index_str]['K'] == 0.25
            assert occupancies[a_index_str]['Na'] == 0.75
        else:
            assert len(occupancies[a_index_str]) == 1
        if a.symbol == 'Cl':
            assert occupancies[a_index_str]['Cl'] == 0.3


content = """
data_1


_chemical_name_common                  'Mysterious something'
_cell_length_a                         5.50000
_cell_length_b                         5.50000
_cell_length_c                         5.50000
_cell_angle_alpha                      90
_cell_angle_beta                       90
_cell_angle_gamma                      90
_space_group_name_H-M_alt              'F m -3 m'
_space_group_IT_number                 225

loop_
_space_group_symop_operation_xyz
   'x, y, z'
   '-x, -y, -z'
   '-x, -y, z'
   'x, y, -z'
   '-x, y, -z'  # some comment
   'x, -y, z'
   'x, -y, -z'
   '-x, y, z'
   'z, x, y'
   '-z, -x, -y'
   'z, -x, -y'
   '-z, x, y'
   '-z, -x, y'
   'z, x, -y'
   '-z, x, -y'
   'z, -x, y'
   'y, z, x'
   '-y, -z, -x'
   '-y, z, -x'
   'y, -z, x'
   'y, -z, -x'
   '-y, z, x'
   '-y, -z, x'
   'y, z, -x'
   'y, x, -z'
   '-y, -x, z'
   '-y, -x, -z'
   'y, x, z'
   'y, -x, z'
   '-y, x, -z'
   '-y, x, z'
   'y, -x, -z'
   'x, z, -y'
   '-x, -z, y'
   '-x, z, y'
   'x, -z, -y'
   '-x, -z, -y'
   'x, z, y'
   'x, -z, y'
   '-x, z, -y'
   'z, y, -x'
   '-z, -y, x'
   'z, -y, x'
   '-z, y, -x'
   '-z, y, x'
   'z, -y, -x'
   '-z, -y, -x'
   'z, y, x'
   'x, y+1/2, z+1/2'
   '-x, -y+1/2, -z+1/2'
   '-x, -y+1/2, z+1/2'
   'x, y+1/2, -z+1/2'
   '-x, y+1/2, -z+1/2'
   'x, -y+1/2, z+1/2'
   'x, -y+1/2, -z+1/2'
   '-x, y+1/2, z+1/2'
   'z, x+1/2, y+1/2'
   '-z, -x+1/2, -y+1/2'
   'z, -x+1/2, -y+1/2'
   '-z, x+1/2, y+1/2'
   '-z, -x+1/2, y+1/2'
   'z, x+1/2, -y+1/2'
   '-z, x+1/2, -y+1/2'
   'z, -x+1/2, y+1/2'
   'y, z+1/2, x+1/2'
   '-y, -z+1/2, -x+1/2'
   '-y, z+1/2, -x+1/2'
   'y, -z+1/2, x+1/2'
   'y, -z+1/2, -x+1/2'
   '-y, z+1/2, x+1/2'
   '-y, -z+1/2, x+1/2'
   'y, z+1/2, -x+1/2'
   'y, x+1/2, -z+1/2'
   '-y, -x+1/2, z+1/2'
   '-y, -x+1/2, -z+1/2'
   'y, x+1/2, z+1/2'
   'y, -x+1/2, z+1/2'
   '-y, x+1/2, -z+1/2'
   '-y, x+1/2, z+1/2'
   'y, -x+1/2, -z+1/2'
   'x, z+1/2, -y+1/2'
   '-x, -z+1/2, y+1/2'
   '-x, z+1/2, y+1/2'
   'x, -z+1/2, -y+1/2'
   '-x, -z+1/2, -y+1/2'
   'x, z+1/2, y+1/2'
   'x, -z+1/2, y+1/2'
   '-x, z+1/2, -y+1/2'
   'z, y+1/2, -x+1/2'
   '-z, -y+1/2, x+1/2'
   'z, -y+1/2, x+1/2'
   '-z, y+1/2, -x+1/2'
   '-z, y+1/2, x+1/2'
   'z, -y+1/2, -x+1/2'
   '-z, -y+1/2, -x+1/2'
   'z, y+1/2, x+1/2'
   'x+1/2, y, z+1/2'
   '-x+1/2, -y, -z+1/2'
   '-x+1/2, -y, z+1/2'
   'x+1/2, y, -z+1/2'
   '-x+1/2, y, -z+1/2'
   'x+1/2, -y, z+1/2'
   'x+1/2, -y, -z+1/2'
   '-x+1/2, y, z+1/2'
   'z+1/2, x, y+1/2'
   '-z+1/2, -x, -y+1/2'
   'z+1/2, -x, -y+1/2'
   '-z+1/2, x, y+1/2'
   '-z+1/2, -x, y+1/2'
   'z+1/2, x, -y+1/2'
   '-z+1/2, x, -y+1/2'
   'z+1/2, -x, y+1/2'
   'y+1/2, z, x+1/2'
   '-y+1/2, -z, -x+1/2'
   '-y+1/2, z, -x+1/2'
   'y+1/2, -z, x+1/2'
   'y+1/2, -z, -x+1/2'
   '-y+1/2, z, x+1/2'
   '-y+1/2, -z, x+1/2'
   'y+1/2, z, -x+1/2'
   'y+1/2, x, -z+1/2'
   '-y+1/2, -x, z+1/2'
   '-y+1/2, -x, -z+1/2'
   'y+1/2, x, z+1/2'
   'y+1/2, -x, z+1/2'
   '-y+1/2, x, -z+1/2'
   '-y+1/2, x, z+1/2'
   'y+1/2, -x, -z+1/2'
   'x+1/2, z, -y+1/2'
   '-x+1/2, -z, y+1/2'
   '-x+1/2, z, y+1/2'
   'x+1/2, -z, -y+1/2'
   '-x+1/2, -z, -y+1/2'
   'x+1/2, z, y+1/2'
   'x+1/2, -z, y+1/2'
   '-x+1/2, z, -y+1/2'
   'z+1/2, y, -x+1/2'
   '-z+1/2, -y, x+1/2'
   'z+1/2, -y, x+1/2'
   '-z+1/2, y, -x+1/2'
   '-z+1/2, y, x+1/2'
   'z+1/2, -y, -x+1/2'
   '-z+1/2, -y, -x+1/2'
   'z+1/2, y, x+1/2'
   'x+1/2, y+1/2, z'
   '-x+1/2, -y+1/2, -z'
   '-x+1/2, -y+1/2, z'
   'x+1/2, y+1/2, -z'
   '-x+1/2, y+1/2, -z'
   'x+1/2, -y+1/2, z'
   'x+1/2, -y+1/2, -z'
   '-x+1/2, y+1/2, z'
   'z+1/2, x+1/2, y'
   '-z+1/2, -x+1/2, -y'
   'z+1/2, -x+1/2, -y'
   '-z+1/2, x+1/2, y'
   '-z+1/2, -x+1/2, y'
   'z+1/2, x+1/2, -y'
   '-z+1/2, x+1/2, -y'
   'z+1/2, -x+1/2, y'
   'y+1/2, z+1/2, x'
   '-y+1/2, -z+1/2, -x'
   '-y+1/2, z+1/2, -x'
   'y+1/2, -z+1/2, x'
   'y+1/2, -z+1/2, -x'
   '-y+1/2, z+1/2, x'
   '-y+1/2, -z+1/2, x'
   'y+1/2, z+1/2, -x'
   'y+1/2, x+1/2, -z'
   '-y+1/2, -x+1/2, z'
   '-y+1/2, -x+1/2, -z'
   'y+1/2, x+1/2, z'
   'y+1/2, -x+1/2, z'
   '-y+1/2, x+1/2, -z'
   '-y+1/2, x+1/2, z'
   'y+1/2, -x+1/2, -z'
   'x+1/2, z+1/2, -y'
   '-x+1/2, -z+1/2, y'
   '-x+1/2, z+1/2, y'
   'x+1/2, -z+1/2, -y'
   '-x+1/2, -z+1/2, -y'
   'x+1/2, z+1/2, y'
   'x+1/2, -z+1/2, y'
   '-x+1/2, z+1/2, -y'
   'z+1/2, y+1/2, -x'
   '-z+1/2, -y+1/2, x'
   'z+1/2, -y+1/2, x'
   '-z+1/2, y+1/2, -x'
   '-z+1/2, y+1/2, x'
   'z+1/2, -y+1/2, -x'
   '-z+1/2, -y+1/2, -x'
   'z+1/2, y+1/2, x'

loop_
   _atom_site_label
   _atom_site_occupancy
   _atom_site_fract_x
   _atom_site_fract_y
   _atom_site_fract_z
   _atom_site_adp_type
   _atom_site_B_iso_or_equiv
   _atom_site_type_symbol
   Na         0.7500  0.000000      0.000000      0.000000     Biso  1.000000 Na
   K          0.2500  0.000000      0.000000      0.000000     Biso  1.000000 K
   Cl         0.3000  0.500000      0.500000      0.500000     Biso  1.000000 Cl
   I          0.5000  0.250000      0.250000      0.250000     Biso  1.000000 I
"""


def test_cif():
    cif_file = io.StringIO(content)

    # legacy behavior is to not read the K atoms
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        atoms_leg = read_cif(cif_file, fractional_occupancies=False)
    elements = np.unique(atoms_leg.get_atomic_numbers())
    for n in (11, 17, 53):
        assert n in elements
    try:
        atoms_leg.info['occupancy']
        raise AssertionError
    except KeyError:
        pass

    cif_file = io.StringIO(content)
    # new behavior is to still not read the K atoms, but build info
    atoms = read_cif(cif_file, fractional_occupancies=True)

    # yield the same old atoms for fractional_occupancies case
    assert len(atoms_leg) == len(atoms)
    assert np.all(atoms_leg.get_atomic_numbers() == atoms.get_atomic_numbers())
    assert atoms_leg == atoms

    elements = np.unique(atoms_leg.get_atomic_numbers())
    for n in (11, 17, 53):
        assert n in elements

    check_fractional_occupancies(atoms)

    # read/write
    fname = 'testfile.cif'
    with open(fname, 'wb') as fd:
        write_cif(fd, atoms)

    with open(fname) as fd:
        atoms = read_cif(fd, fractional_occupancies=True)

    check_fractional_occupancies(atoms)

    # check repeating atoms
    atoms = atoms.repeat([2, 1, 1])
    assert len(atoms.arrays['spacegroup_kinds']) == len(atoms.arrays['numbers'])


# ICSD-like file from issue #293
content2 = """
data_global
_cell_length_a 9.378(5)
_cell_length_b 7.488(5)
_cell_length_c 6.513(5)
_cell_angle_alpha 90.
_cell_angle_beta 91.15(5)
_cell_angle_gamma 90.
_cell_volume 457.27
_cell_formula_units_Z 2
_symmetry_space_group_name_H-M 'P 1 n 1'
_symmetry_Int_Tables_number 7
_refine_ls_R_factor_all 0.071
loop_
_symmetry_equiv_pos_site_id
_symmetry_equiv_pos_as_xyz
1 'x+1/2, -y, z+1/2'
2 'x, y, z'
loop_
_atom_type_symbol
_atom_type_oxidation_number
Sn2+ 2
As4+ 4
Se2- -2
loop_
_atom_site_label
_atom_site_type_symbol
_atom_site_symmetry_multiplicity
_atom_site_Wyckoff_symbol
_atom_site_fract_x
_atom_site_fract_y
_atom_site_fract_z
_atom_site_B_iso_or_equiv
_atom_site_occupancy
_atom_site_attached_hydrogens
Sn1 Sn2+ 2 a 0.5270(2) 0.3856(2) 0.7224(3) 0.0266(4) 1. 0
Sn2 Sn2+ 2 a 0.0279(2) 0.1245(2) 0.7870(2) 0.0209(4) 1. 0
As1 As4+ 2 a 0.6836(4) 0.1608(5) 0.8108(6) 0.0067(7) 1. 0
As2 As4+ 2 a 0.8174(4) 0.6447(6) 0.1908(6) 0.0057(6) 1. 0
Se1 Se2- 2 a 0.4898(4) 0.7511(6) 0.8491(6) 0.0110(6) 1. 0
Se2 Se2- 2 a 0.7788(4) 0.6462(6) 0.2750(6) 0.0097(6) 1. 0
Se3 Se2- 2 a 0.6942(4) 0.0517(5) 0.5921(6) 0.2095(6) 1. 0
Se4 Se2- 2 a 0.0149(4) 0.3437(6) 0.5497(7) 0.1123(7) 1. 0
Se5 Se2- 2 a 0.1147(4) 0.5633(4) 0.3288(6) 0.1078(6) 1. 0
Se6 Se2- 2 a 0.0050(4) 0.4480(6) 0.9025(6) 0.9102(6) 1. 0
"""


def test_cif_icsd():
    cif_file = io.StringIO(content2)
    atoms = read_cif(cif_file)
    # test something random so atoms is not unused
    assert 'occupancy' in atoms.info


@pytest.fixture()
def cif_atoms():
    cif_file = io.StringIO(content)
    return read_cif(cif_file)


def test_cif_loop_keys(cif_atoms):
    data = {}
    # test case has 20 entries
    data['someKey'] = [[str(i) + "test" for i in range(20)]]
    # test case has 20 entries
    data['someIntKey'] = [[str(i) + "123" for i in range(20)]]
    cif_atoms.write('testfile.cif', loop_keys=data)

    atoms1 = read_cif('testfile.cif', store_tags=True)
    # keys are read lowercase only
    r_data = {'someKey': atoms1.info['_somekey'],
              'someIntKey': atoms1.info['_someintkey']}
    assert r_data['someKey'] == data['someKey'][0]
    # data reading auto converts strins
    assert r_data['someIntKey'] == [int(x) for x in data['someIntKey'][0]]


# test if automatic numbers written after elements are correct
def test_cif_writer_label_numbers(cif_atoms):
    cif_atoms.write('testfile.cif')
    atoms1 = read_cif('testfile.cif', store_tags=True)
    labels = atoms1.info['_atom_site_label']
    # cannot use atoms.symbols as K is missing there
    elements = atoms1.info['_atom_site_type_symbol']
    build_labels = [
        "{:}{:}".format(
            x, i) for x in set(elements) for i in range(
            1, elements.count(x) + 1)]
    assert build_labels.sort() == labels.sort()


def test_cif_labels(cif_atoms):
    data = [["label" + str(i) for i in range(20)]]  # test case has 20 entries
    cif_atoms.write('testfile.cif', labels=data)

    atoms1 = read_cif('testfile.cif', store_tags=True)
    print(atoms1.info)
    assert data[0] == atoms1.info['_atom_site_label']


def test_cifloop():
    dct = {'_eggs': range(4),
           '_potatoes': [1.3, 7.1, -1, 0]}

    loop = CIFLoop()
    loop.add('_eggs', dct['_eggs'], '{:<2d}')
    loop.add('_potatoes', dct['_potatoes'], '{:.4f}')

    string = loop.tostring() + '\n\n'
    lines = string.splitlines()[::-1]
    assert lines.pop() == 'loop_'

    newdct = parse_loop(lines)
    print(newdct)
    assert set(dct) == set(newdct)
    for name in dct:
        assert dct[name] == pytest.approx(newdct[name])


@pytest.mark.parametrize('data', [b'', b'data_dummy'])
def test_empty_or_atomless(data):
    ciffile = io.BytesIO(data)

    images = read_cif(ciffile, index=':')
    assert len(images) == 0


def test_empty_or_atomless_cifblock():
    blocks = parse_string('data_dummy')

    assert len(blocks) == 1
    assert not blocks[0].has_structure()
    with pytest.raises(NoStructureData):
        blocks[0].get_atoms()


def test_symbols_questionmark():
    blocks = parse_string("""\
data_dummy
loop_
_atom_site_label
?""")

    assert not blocks[0].has_structure()
    with pytest.raises(NoStructureData, match='undetermined'):
        blocks[0].get_atoms()


def test_bad_occupancies(cif_atoms):
    assert 'Au' not in cif_atoms.symbols
    cif_atoms.symbols[0] = 'Au'
    with pytest.warns(UserWarning, match='no occupancy info'):
        write_cif('tmp.cif', cif_atoms)


@pytest.mark.parametrize(
    'setting_name, ref_setting',
    [
        ('hexagonal', 1),
        ('trigonal', 2),
        ('rhombohedral', 2)
    ]
)
def test_spacegroup_named_setting(setting_name, ref_setting):
    """The rhombohedral crystal system signifies setting=2"""
    blocks = parse_string("""\
data_test
_space_group_crystal_system {}
_symmetry_space_group_name_H-M         'R-3m'
""".format(setting_name))

    assert len(blocks) == 1
    spg = blocks[0].get_spacegroup(False)
    assert int(spg) == 166
    assert spg.setting == ref_setting


@pytest.fixture()
def atoms():
    return Atoms('CO', cell=[2., 3., 4., 50., 60., 70.], pbc=True,
                 scaled_positions=[[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]])


def roundtrip(atoms):
    buf = to_bytes(atoms, format='cif')
    return parse_atoms(buf, format='cif')


def test_cif_roundtrip_periodic(atoms):
    # Reading and writing the cell loses the rotation information,
    # but preserves cellpar and scaled positions.
    atoms1 = roundtrip(atoms)

    assert str(atoms1.symbols) == 'CO'
    assert all(atoms1.pbc)
    assert atoms.cell.cellpar() == pytest.approx(
        atoms1.cell.cellpar(), abs=1e-5)
    assert atoms.get_scaled_positions() == pytest.approx(
        atoms1.get_scaled_positions(), abs=1e-5)


def test_cif_roundtrip_nonperiodic():
    atoms = molecule('H2O')
    atoms1 = roundtrip(atoms)
    assert not compare_atoms(atoms, atoms1, tol=1e-5)


def test_cif_missingvector(atoms):
    # We don't know any way to represent only 2 cell vectors in CIF.
    # So we discard them and warn the user.
    atoms.cell[0] = 0.0
    atoms.pbc[0] = False

    assert atoms.cell.rank == 2

    with pytest.raises(ValueError, match='CIF format can only'):
        roundtrip(atoms)


def test_cif_roundtrip_mixed():
    atoms = Atoms('Au', cell=[1., 2., 3.], pbc=[1, 1, 0])
    atoms1 = roundtrip(atoms)

    # We cannot preserve PBC info for this case:
    assert all(atoms1.pbc)
    assert compare_atoms(atoms, atoms1, tol=1e-5) == ['pbc']
    assert atoms.get_scaled_positions() == pytest.approx(
        atoms1.get_scaled_positions(), abs=1e-5)


def test_loop_with_space():
    # Regression test for https://gitlab.com/ase/ase/-/issues/859 .
    # (We do the \n\ to avoid automatic trailing whitespace cleanup)
    blocks = parse_string("""\
data_image0
loop_
 _hello
 banana
 \n\
_potato 42
""")

    assert len(blocks) == 1
    assert blocks[0]['_potato'] == 42


# Test case for symmetry loop without any symops in it.
empty_symloop_cif = """
data_cif

_audit_creation_method             'generated by GULP'

_symmetry_space_group_name_H-M     '(unknown)       '
_symmetry_Int_Tables_number        1

loop_
_symmetry_equiv_pos_site_id
_symmetry_equiv_pos_as_xyz
_cell_length_a                        5.1053
_cell_length_b                        5.1053
_cell_length_c                        5.1053
_cell_angle_alpha                   60.0000
_cell_angle_beta                    60.0000
_cell_angle_gamma                   60.0000

loop_
_atom_site_label
_atom_site_fract_x
_atom_site_fract_y
_atom_site_fract_z
_atom_site_occupancy
  Cu          0.00000      0.00000      0.00000  1.0000
  Cu          0.00000      0.00000      0.50000  1.0000
  Cu          0.00000      0.50000      0.00000  1.0000
  Cu          1.00000      0.50000      0.50000  1.0000
  Cu          0.50000      0.00000      0.00000  1.0000
  Cu          0.50000      1.00000      0.50000  1.0000
  Cu          0.50000      0.50000      0.00000  1.0000
  Cu          0.50000      0.50000      0.50000  1.0000
"""


def test_cif_with_empty_symmetries_loop():
    blocks = parse_string(empty_symloop_cif)
    assert len(blocks) == 1
    atoms = blocks[0].get_atoms()
    assert str(atoms.symbols) == 'Cu8'
    assert atoms.cell.lengths() == pytest.approx(5.1053)
    assert atoms.cell.angles() == pytest.approx(60)


def test_cif_with_strings_for_integers():
    # Test that even if spacegroup is given as '42' with ticks, we can
    # still convert it to integer and thus actual Spacegroup.

    buf = """\
data_cif
_cell_length_a               1
_cell_length_b               1
_cell_length_c               1
_cell_angle_alpha            90
_cell_angle_beta             90
_cell_angle_gamma            90
_symmetry_space_group_H-M        'P1'
_symmetry_Int_Tables_number      '42'
"""

    block = parse_string(buf)[0]
    assert block.get_spacegroup(subtrans_included=True).no == 42



# In this script the Virtual Crystal approximation is used to model
# a stronger affinity for positive charge on the H atoms.
# This could model interaction with other molecules not explicitly
# handled.
import numpy as np

from ase import Atoms
from ase.calculators.siesta import Siesta
from ase.calculators.siesta.parameters import Species
from ase.optimize import QuasiNewton

atoms = Atoms('CH4', np.array([
    [0.000000, 0.000000, 0.000000],
    [0.682793, 0.682793, 0.682793],
    [-0.682793, -0.682793, 0.682790],
    [-0.682793, 0.682793, -0.682793],
    [0.682793, -0.682793, -0.682793]]))

siesta = Siesta(
    species=[
        Species(symbol='H', excess_charge=0.1)])

atoms.calc = siesta
dyn = QuasiNewton(atoms, trajectory='h.traj')
dyn.run(fmax=0.02)



"""Tests for FixCom."""
import numpy as np
import pytest

from ase import Atoms
from ase.build import molecule
from ase.calculators.emt import EMT
from ase.constraints import FixCom
from ase.md.velocitydistribution import MaxwellBoltzmannDistribution
from ase.optimize import BFGS


@pytest.fixture(name="atoms")
def fixture_atoms() -> Atoms:
    """fixture_atoms"""
    atoms = molecule('H2O')
    atoms.center(vacuum=4)
    atoms.calc = EMT()
    return atoms


@pytest.mark.optimize()
def test_center_of_mass_position(atoms: Atoms):
    """Test if the center of mass does not move."""
    cold = atoms.get_center_of_mass()
    atoms.set_constraint(FixCom())

    assert atoms.get_number_of_degrees_of_freedom() == 6

    with BFGS(atoms) as opt:
        opt.run(steps=5)

    cnew = atoms.get_center_of_mass()

    assert max(cnew - cold) == pytest.approx(0.0, abs=1e-8)


@pytest.mark.optimize()
def test_center_of_mass_velocity(atoms: Atoms):
    """Test if the center-of-mass veloeicty is zero."""
    atoms.set_constraint(FixCom())

    # `adjust_momenta` of constaints are applied inside
    MaxwellBoltzmannDistribution(atoms, temperature_K=300.0)

    velocity_com = atoms.get_momenta().sum(axis=0) / atoms.get_masses().sum()

    assert max(velocity_com) == pytest.approx(0.0, abs=1e-8)


@pytest.mark.optimize()
def test_center_of_mass_force(atoms: Atoms):
    """Test if the corrected forces are along the COM-preserving direction."""
    rnd = np.random.default_rng(42)
    forces = rnd.random(size=atoms.positions.shape)
    FixCom().adjust_forces(atoms, forces)
    np.testing.assert_allclose(atoms.get_masses() @ forces, 0.0, atol=1e-12)
